# This class implements a listing object for Asm objects to share as they translate .asm files
# into the .lst file built and stored here.

from MyTypes import filelist

class AsmList(filelist):
    """Formats listing lines and appends them to self.  The format is:
    line# addr opercode hexicode long___label original  line    $from   ;.asm file"""
    # parts = (oldPC, label, opcode, hexop, mne, opfield, line)

    def AddLine(self, linenum, parts, printlist):
        pc, label, opcode, hexop, mne, opfield, line = parts
        try:
            pc = hex(pc)[2:].upper().zfill(4)
        except (ValueError, TypeError):
            pc = '    '

        self.append(' '.join([str(linenum).rjust(5), pc,
                              opcode.rjust(8), hexop.ljust(8),
                              label.ljust(12), mne.ljust(6),
                               '%s   %s' %(opfield, line)]) + '\n')
        if printlist:
            print self[linenum - 1],

    def FixLine(self, i, size, value):
        self[i] = self[i].replace('??' * size, hex(value)[2:].upper().zfill(2 * size), 1)

    def GetMachCode(self, linenum):
        addr, opcode, hexop = self[linenum][6:10], self[linenum][11:19], self[linenum][20:28]
        if opcode.isspace():
            return None, None
        return int(addr, 16), opcode.lstrip() + hexop.rstrip()
