from BC11AsmClass import Asm
from AsmList import AsmList
from SRecord import SRecord
from progressDialog import progressDialog
import re

class Assembler:

    def __init__(self, editWindow):
        self.edWin = editWindow
        self.listing = AsmList()        # Filled in by instance(s) of Asm
        self.asm = Asm(self.listing)    # Create an instance of Asm
        self.errors = 0

    def WriteList(self, fname):
        out = open(fname, 'w')
        out.writelines(self.listing)
        out.close()

    def WriteXref(self, fname, xref):
        out = open(fname, 'w')
        out.writelines(xref)
        out.close()

    def WriteSrec(self, fname, srec):
        out = open(fname, 'w')
        out.writelines(srec)
        out.close()

    def MakeSym(self, fname):
        labels = Asm.labels
        out = open(fname, 'w')
        for label in labels:
            out.write(' '.join([label, hex(labels[label])[2:].upper().zfill(4)]) + '\n')
        out.close()

    def MakeS19(self):
        listing = self.listing
        s19 = SRecord()            # Create an SRecord instance
        bytecount = 0
        for i, line in enumerate(listing):
            addr, hexdat = listing.GetMachCode(i)
            if addr:
                s19.AddBytes(addr, hexdat)
                bytecount += len(hexdat)/2
        s19.EndS19()
        return s19, bytecount

    def PrintErrors(self):
        fwds = Asm.forwards
        print
        print 'Errors due to unresolved forward references'
        print 'Name'.ljust(12), 'Line #'.ljust(6), 'Addm','FROption'
        print '----'.ljust(12), '------'.ljust(6), '----','--------'
        for name in fwds:
            for tuplist in fwds[name]:
                errorline = ' '.join([name.ljust(12),
                                     str(tuplist[0] + 1).ljust(6),
                                     str(tuplist[1]).ljust(4), ''])
                if type(tuplist[2]) == int:
                    errorline += '$' + hex(tuplist[2])[2:].upper()
                elif type(tuplist[2]) == str:
                    errorline += tuplist[2]
                else:
                    errorline += '????'
                errorline += '\n'
                print errorline,
        print

    def CrossRef(self):
        result = []
        count = len(Asm.listing)
        listing = ''.join(Asm.listing)
        del Asm.listing
        prelabel = r'[\s\!\+\-\*\/\&\|\#]?('
        postlabel = r')[\s\+\-\*\/\&\|]+'
        MakePLHeader(result)
        labels = Asm.labels
        if labels:
            labelkeys = sorted(labels.keys())
            if count > 200:
                labelProgress = progressDialog(self.edWin.top, 'Cross Referencing Labels')
                labelProgress.SetMax(len(labels))
                for i, label in enumerate(labelkeys):
                    canceled = labelProgress.Run(i)
                    if canceled:
                        print 'Cross Referencing Labels canceled'
                        result.append('Canceled')
                        break
                    xref = GetLineNums(prelabel + label + postlabel, listing)
                    xrefs = MakeLines(xref)
                    result.append(MakeLine(label, labels[label], xrefs))
                if not canceled:
                    labelProgress.close()
            else:
                self.edWin.top.config(cursor="watch")
                for label in labelkeys:
#                    self.edWin.top.update
                    self.edWin.root.update()
                    xref = GetLineNums(prelabel + label + postlabel, listing)
                    xrefs = MakeLines(xref)
                    result.append(MakeLine(label, labels[label], xrefs))
        else:
            result.append('None')
        MakeDCHeader(result)
        constants = Asm.constants
        if constants:
            constkeys = sorted(constants.keys())
            if count > 200:
                constProgress = progressDialog(self.edWin.top, 'Cross Referencing Constants')
#                self.edWin.top.update_idletasks()
                constProgress.SetMax(len(constants))
                for i, name in enumerate(constkeys):
                    canceled = constProgress.Run(i)
                    if canceled:
                        print 'Cross Referencing Constants canceled'
                        result.append('Canceled')
                        break
                    xref = GetLineNums(prelabel + name + postlabel, listing)
                    xrefs = MakeLines(xref)
                    result.append(MakeLine(name, constants[name], xrefs))
                if not canceled:
                    constProgress.close()
            else:
                self.edWin.top.config(cursor="watch")
                for name in constkeys:
                    self.edWin.root.update()
                    xref = GetLineNums(prelabel + name + postlabel, listing)
                    xrefs = MakeLines(xref)
                    result.append(MakeLine(name, constants[name], xrefs))
        else:
            result.append('None')
        self.edWin.top.config(cursor="")
        return result



    def assemble(self, asmfilename):
        self.asmfilename = asmfilename
        self.errors = self.asm.AsmFile(asmfilename) # Run Asm

    def reset(self):
        self.listing.reset()
        self.asm.reset()

## Helpers of CrossReference ##
def GetLineNums(repattern, listing):
    xref = []
    i = re.finditer(repattern, listing)
    for match in i:
        a = match.start(1)
        if a < 0:
            print 'bad match'
            break
        linestart = listing.rfind('\n', a - 100, a) + 1
        linenum = listing[linestart:linestart + 5].lstrip()
        if a - linestart == 29:
            xref.insert(0,linenum)
        else:
            xref.append(linenum)
    return xref

def MakeLine(key, value, xrefs):
    return ' '.join([key.ljust(12), '$' +
                     hex(value)[2:].upper().zfill(4), xrefs]) + '\n'
        
def MakeLines(xref):
    xlines = len(xref)
    xrefs = ', '.join(xref[:12])
    if xlines > 12:
        xrefs += ',\n' + ' ' * 19 + ', '.join(xref[12:24])
    if xlines > 24:
        xrefs += ',\n' + ' ' * 19 + ', '.join(xref[24:36])
    if xlines > 36:
        xrefs += ',\n' + ' ' * 19 + ', '.join(xref[36:48])
    if xlines > 48:
        xrefs += ',\n' + ' ' * 19 + ', '.join(xref[48:])
    return xrefs

def MakePLHeader(result):
    proglabline = ' Program Labels cross reference:\n'
    result.append(proglabline)
    proglabline = ' '.join([' Label Name ', 'Addrs', 'Lines (defined, referenced, ...)']) + '\n'
    result.append(proglabline)
    proglabline = ' '.join(['-' * 12, '-' * 5, '-' * 33]) + '\n'
    result.append(proglabline)

def MakeDCHeader(result):
    result.append('\n')
    proglabline = ' Declared Constants cross reference:\n'
    result.append(proglabline)
    proglabline = ' '.join([' Const Name ', 'Value', 'Lines (defined, referenced, ...)']) + '\n'
    result.append(proglabline)
    proglabline = ' '.join(['-' * 12, '-' * 5, '-' * 33]) + '\n'
    result.append(proglabline)

if __name__ == '__main__':
    import os
#    pwd = os.getcwd()
#    os.chdir(pwd + '\\EGT')
    filename = os.getcwd() + '\\test.asm'
    assembler = Assembler(filename)
    assembler.assemble()

    if assembler.asm.forwards:
        assembler.PrintErrors()
    elif not assembler.errors:
        print
        head, tail = os.path.split(filename)
        base, ext = os.path.splitext(tail)
        listname = base + '.lst'
        xrefname = base + '.xrf'
        srecname = base + '.s19'
        symname = base + '.sym'
        assembler.WriteList(listname, assembler.listing)
        xref = assembler.CrossRef()
        assembler.WriteXref(xrefname, xref)
        s19 = assembler.MakeS19()
        assembler.WriteSrec(srecname, s19)
        assembler.MakeSym(symname)
        for line in xref:
            print line,

