import os
import os.path
import sys

# I Want Python to search for replaced parts of IDLE here first
try:
    a = sys.path.index('C:\\bc11asm')
    sys.path.pop(a)
except ValueError:
    pass
sys.path.insert(0,'C:\\bc11asm')

import exceptions
from Tkinter import *
import tkMessageBox
from time import time
from EditorWindow import EditorWindow, fixwordbreaks
from FileList import FileList
from OutputWindow import OutputWindow
from configHandler import bc11Conf

from Assembler import Assembler
import debugDialog

class BC11IDEditorWindow(EditorWindow):
    """Regular EditorWindow window from IDLE, but it supports assembly
    of file which have a .asm file extention and conform to the BCDasm rules,
    and has a graphical interface to Motorola's EVB running BUFFALO."""

    def __init__(self, *args):
        EditorWindow.__init__(self, *args)
        self.top.bind("<<assemble-file>>", self.assemble_now)
        self.top.bind("<<run-assembly>>", self.open_debugger)
        self.assembler = Assembler(self)   # Create the Assembler

    def open_debugger(self, event=None):
        if not self.io.get_saved():
            self.io.save(None)
        fname = self.io.filename
        if fname.endswith('.asm') or fname.endswith('.lst'):
            try:
                asmtime = os.stat(fname[:-4] + '.asm')[8]   # .asm modification date
                lsttime = os.stat(fname[:-4] + '.lst')[8]   # .lst modification date
                xrefname = fname[:-4] + '.xrf'
                xrftime = os.stat(xrefname)[8]              # .xrf modification date
            except OSError:
                message = 'A file error occurred. Had read ' + fname or 'nothing'
                tkMessageBox.showwarning(message=message)
                return
            if asmtime > lsttime:
                message = 'The .lst file is older than the .asm file.\Assembly now?'
                if tkMessageBox.askyesno(message=message):
                    self.assemble_now()
                    xrftime = os.stat(xrefname)[8]          # .xrf modification date
                else:
                    return
            if asmtime > xrftime:
                message = 'The .xrf file is older than the .asm file.\nCross Reference now?'
                if tkMessageBox.askyesno(message=message):
                    self.crossref(xrefname)
                    self.assembler.reset()          # Reset the Assembler
                else:
                    return
            if self.flist.debugger:
                dbug = self.flist.debugger
                dbug.top.wakeup()
                dbug.s19value.set(fname[:-4] + '.s19')
                dbug.top.after_idle(dbug.Load)
            else:
                self.flist.debugger = debugDialog.Debugger(self.root, menu=self.menubar,
                                                           flist=self.flist,
                                                           srecord=fname[:-4] + '.s19')

    def assemble_now(self, event=None):
        self.flist.open_shell()             # Put the shell window on top for assembler messages
        if not self.io.get_saved():
            self.io.save(None)
        filename = self.io.filename         # Get the name of the file for this window
        head, tail = os.path.split(filename)    # Find out which directory we need to be in
        base, ext = os.path.splitext(tail)
        if ext.lower() == '.asm':
            print 'Assembling'
            cwd = os.getcwd()
            os.chdir(head)                      # Change to that directory
            listname = base + '.lst'            # use these file extentions
            xrefname = base + '.xrf'
            srecname = base + '.s19'
            starttime = time()
            self.assembler.assemble(filename)           # Run the assembly
            endtime = time() - starttime
            print 'Assembly time: %4.3f seconds' % endtime
            if self.assembler.asm.forwards:
                self.assembler.PrintErrors()    # Print errors, if any
            elif not self.assembler.errors:     # If not
                self.assembler.WriteList(listname)   # Write the listing to disk
                s19, bytecount = self.assembler.MakeS19()       # Create the S record
                bytereport = '\nByte Count: %3.2fKB (%d bytes).\n' %(bytecount/1024.0, bytecount)
                print bytereport
                self.assembler.WriteSrec(srecname, s19)              # Write S record to disk
                listing = file(listname, mode='a')
                listing.write(bytereport)
                listing.close()
                if tkMessageBox.askyesno(message='Cross Reference now?'):
                    self.crossref(xrefname)
                print 'Done!'
                self.flist.open(listname)       # Open the listing
#                self.flist.open(xrefname)       # Open it
            if event:
                self.assembler.reset()          # Reset the Assembler
            os.chdir(cwd)                   # Set the directory back to what it was when we started

    def crossref(self, xrefname):
        print 'Cross Referencing...'
        starttime = time()
        xref = self.assembler.CrossRef()     # Create the cross reference
        self.assembler.WriteXref(xrefname, xref)         # Write cross reference to disk
        endtime = time() - starttime
        print 'Cross Reference time: %4.3f seconds' % endtime

    def _close(self):
        EditorWindow._close(self)


class BC11IDEFileList(FileList):
    """Extend base class: Add open_shell and
    override write() with tkconsole.stderr.write()"""

    # override FileList's class variable, instances return BC11IDEditorWindow
    # instead of EditorWindow when new edit windows are created.
    EditorWindow = BC11IDEditorWindow

    debugger = None
    myIDEShell = None

    def open_shell(self, event=None):
        if self.myIDEShell:
            self.myIDEShell.top.wakeup()
        else:
            self.myIDEShell = BC11IDEShell(self)
            if self.myIDEShell:
                if not self.myIDEShell.begin():
                    return None
        return self.myIDEShell

    def write(self, s):
        "Override base class method"
        self.tkconsole.stderr.write(s)

class ModifiedInterpreter:

    def __init__(self, tkconsole):
        self.tkconsole = tkconsole
        locals = sys.modules['__main__'].__dict__


class BC11IDEShell(OutputWindow):    #Use to send as TK's tkconsole

    # Override menus
    menu_specs = [
        ("file", "_File"),
        ("edit", "_Edit"),
        ("debug", "_Debug"),
        ("options", "_Options"),
        ("windows", "_Windows"),
        ("help", "_Help"),
    ]

    def __init__(self, flist=None):
        self.interp = ModifiedInterpreter(self)
        if flist is None:
            root = Tk()
            fixwordbreaks(root)
            root.withdraw()
            flist = BC11IDEFileList(root)
        #
        OutputWindow.__init__(self, flist, None, None)
        #
        self.save_stdout = sys.stdout
        self.save_stderr = sys.stderr
        self.save_stdin = sys.stdin
        import IOBinding
        self.stdout = PseudoFile(self, "stdout", IOBinding.encoding)
#        self.stderr = PseudoFile(self, "stderr", IOBinding.encoding)
#        self.console = PseudoFile(self, "console", IOBinding.encoding)
        sys.stdout = self.stdout
#        sys.stderr = self.stderr
#        sys.stdin = self.console

    def get_standard_extension_names(self):
        return bc11Conf.GetExtensions(shell_only=True)

    def settitle(self):
        self.top.title("BCD11IDE")
#        self.top.wm_iconname("BCD11IDE")

    def reallyclose(self):
        message = "Do you really want to close all windows and exit"
        m = tkMessageBox.Message(
            title = "Really Exit",
            message = message,
            icon = tkMessageBox.QUESTION,
            type = tkMessageBox.OKCANCEL,
            master = self.text)
        reply = m.show()
        return reply

    def close(self):
        "Extend EditorWindow.close()"
        if self.reallyclose() == "ok":
            flist = self.flist
            for edit in flist.inversedict.keys():
                if edit == self:
                    continue
                reply = edit.close()
                if reply == "cancel":
                    break
            return EditorWindow.close(self)
        self.text.focus_set()

    def _close(self):
        "Extend EditorWindow._close() and shut down debugger"
        # Restore std streams
        sys.stdout = self.save_stdout
        sys.stderr = self.save_stderr
        sys.stdin = self.save_stdin
        # Break cycles
        self.console = None
        self.flist.myIDEShell = None
        self.history = None
        EditorWindow._close(self)

    COPYRIGHT = \
          'Based on IDLE\'s EditorWindow and PyShell.  Go to Python.org for more information.'

    def begin(self):
        self.text.mark_set("iomark", "end-1c")
        print """--- Welcome to BCD11IDE. BCD's Integrated Development Environment for the MC68HC11. ---"""
        print ' ' * 4 + BC11IDEShell.COPYRIGHT + '\n' * 2
        return True

    def write(self, s, tags=()):
        try:
            self.text.mark_set('insert', END)
            OutputWindow.write(self, s, tags)
        except:
            pass

class PseudoFile:

    def __init__(self, shell, tags, encoding=None):
        self.shell = shell
        self.tags = tags
        self.softspace = 0
        self.encoding = encoding

    def write(self, s):
        self.shell.write(s, self.tags)

    def writelines(self, l):
        map(self.write, l)

    def flush(self):
        pass

    def isatty(self):
        return True

def main():
#    global flist, root      # I wonder how KEY this is to making things work ???
    enable_shell = False
    enable_edit = False

    # Get the BCD11IDE settings configuration
    edit_start = bc11Conf.GetOption('main', 'General',
                                    'editor-on-startup', type='bool')
    enable_edit = enable_edit or edit_start
    enable_shell = enable_shell or not edit_start

    # start editor and/or shell windows:
#    root = Tk(className="Idle")
    root = Tk()
    fixwordbreaks(root)
    root.withdraw()     # The Tk root window stays hidden
    flist = BC11IDEFileList(root)
    if enable_edit:     # This is here because config will probably be the
        flist.new()     # last thing that I get around to modifing.
    if enable_shell:
        if not flist.open_shell():
            return      # couldn't open shell
    shell = flist.myIDEShell    # In practice, the shell will ALWAYS open
    shell.settitle()

    # Default directory for opening from IDEShell
    os.chdir(r'C:\bc11asm')
    shell.io.open()
    root.mainloop()
    root.destroy()
if __name__ == "__main__":
    sys.modules['BC11IDEShell'] = sys.modules['__main__']
main()
