from WindowList import ListedToplevel
from progressDialog import progressDialog
from tkSimpleDialog import askinteger
import BCDbaseDialog
import tkFileDialog
import tkMessageBox
from Tkinter import *
from Queue import *
import threading
from time import time, sleep
from win32file import *     # The base COM port and file IO functions.
from win32event import *    # We use events and the WaitFor[Multiple]Objects functions.
import win32con             # constants.


notkeys = set(['Shift_L', 'Shift_R', 'Control_R', 'Control_L', 'Win_L', 'Win_R', 'Prior', 'Next',
               'Alt_L', 'Alt_R', 'Tab', 'Right', 'Left','Up', 'Down', 'Escape', 'Insert', 'Home',
               'F1', 'F2', 'F3', 'F4', 'F5', 'F6', 'F7', 'F8', 'F9', 'F10', 'L1', 'L2', 'End'])

class Debugger:
    """Dialog Window for BCD11IDE's debugger.

    """
    def __init__(self, root, menu=None, flist=None, srecord=''):
        self.root = root
        self.top = top = ListedToplevel(root, menu=menu)
        self.flist = flist

        top.option_add("*Entry*font", "Courier 10", 20)
        top.configure(borderwidth=5)
        top.geometry("+%d+%d" % (root.winfo_rootx()+75,
                                  root.winfo_rooty()+100))
        if menu is None:
            root.withdraw()     # Here, we are running from __main__
        self.bg = "SystemButtonFace"
##        self.bg = "white"
        self.baudrate = baudrate = 9600     # Needed by baudrate value widget
        self.quetime = quetime = 20.0/baudrate + .010

        self.CreateWidgets(top)                     # Fit widgets into self
        top.resizable(height=FALSE, width=FALSE)    # then lock the size

        top.title('BCD11IDE Debugger')
        top.protocol("WM_DELETE_WINDOW", self.close)
        top.tkraise()
        top.wm_deiconify()

        self.s19value.set(srecord)

        self.textView.focus_set()
        top.focused_widget = self.textView

        self.running = True             # Safety for shutting down background threads
        self.queue = queue = Queue()    # Queue for inter-tread communication
        self.tty = SerialTTY(self, "COM1", baudrate)    # Pass queue by reference?
        listener = threading.Thread(target=self.tty.read)
        listener.start()                 # Call tty.read() in a background thread!

        self.labelxref = {}         # Dictionary of labels and their values
        self.inversexref = {}       # Inverse dictionary of labels and their values
        self.constxref = {}         # Dictionary of constants and their values
        self.asmAddrList = []       # List of address in the loaded context

##        self.top.update_idletasks()
##        self.top.wait_visibility() # Make window visible

        self.SendChars('\r')
        reply = self.PollQueue(expect=1130, wait_one=True)
        if not reply:
            print reply
            self.notresponding()
            return
        self.Converse(chr(0x18), replylen=4)
        reply = self.Converse('call e000\r', replylen=80)    # Reset Buffalo
        print >> self, reply
        self.Converse('\r')  # Terminate Buffalo startup msg wait
        sleep(quetime)          # Wait for queue to be filled
        self.PeriodicCall()     # Start background listener
        self.Update(True)       # Fill in register entries
        self.SendCommand('help')    # Since we just called RM, prime 'help' command


#################################


    def CreateWidgets(self, top):
        self.gui_only = False
        self.frameMain = frameMain = Frame(top, borderwidth=2, relief=SUNKEN)
        frameMain.grid(row=0, column=0)
        frameBg = Frame(frameMain)  # Make the background frame
##        frameBg = Frame(frameMain, bg=self.bg)  # Make the background frame
        frameBg.pack(expand=TRUE, fill=BOTH)
        # Put labels across row zero
        Label(frameBg, text = 'Registers', bg=self.bg
              ).grid(row=0, column=0, sticky=EW, columnspan=2, padx=5, pady=5)
        Label(frameBg, text='Addr / Label', bg=self.bg
              ).grid(row=0, column=2, sticky=W, columnspan=2, padx=5, pady=5)
        Label(frameBg, text='Value', bg=self.bg
              ).grid(row=0, column=4, sticky=W, columnspan=1, padx=5, pady=5)
        Label(frameBg, text='Address', bg=self.bg
              ).grid(row=0, column=5, sticky=W, columnspan=1, padx=5, pady=5)
        Label(frameBg, text=' Machine  Code ', bg=self.bg
              ).grid(row=0, column=6, sticky=W, columnspan=1, padx=5, pady=5)
        Label(frameBg, text='Label  Name ', bg=self.bg
              ).grid(row=0, column=7, sticky=W, columnspan=1, padx=5, pady=5)
        Label(frameBg, text='Mnemonic and Operands', bg=self.bg
              ).grid(row=0, column=8, sticky=EW, columnspan=2, padx=5, pady=5)
        Label(frameBg, text='Command', bg=self.bg
              ).grid(row=0, column=11, sticky=W, columnspan=1, padx=5, pady=5)

        # Fill in row 1
        Label(frameBg, text='PC', bg=self.bg
              ).grid(row=1, column=0, sticky=E, columnspan=1, padx=5)
        pcvalue = StringVar(top)
        pcentry = Entry(frameBg, textvariable=pcvalue, width=5)
        pcentry.grid(row=1, column=1, sticky=W, columnspan=1, padx=5)
        l1value = StringVar(top)
        l1entry = Entry(frameBg, textvariable=l1value, width=12)
        l1entry.grid(row=1, column=2, sticky=W, columnspan=2, padx=5)
        m1value = StringVar(top)
        m1entry = Entry(frameBg, textvariable=m1value, width=3)
        m1entry.grid(row=1, column=4, sticky=W, columnspan=1, padx=5)

        frameDisAsm = Frame(frameBg)
        frameDisAsm.grid(row=1, column=5, sticky=N+SW, rowspan=6, columnspan=6, padx=5)
        self.disAsmList = Listbox(frameDisAsm, width=60, height=8, takefocus=FALSE,
                                  font=('Courier', 8), exportselection=FALSE)
        self.dalScroll = Scrollbar(frameDisAsm, command=self.disAsmList.yview)
        self.disAsmList.config(yscrollcommand=self.dalScroll.set)
        self.disAsmList.pack(side=LEFT, expand=TRUE, fill=BOTH)
        self.dalScroll.pack(side=LEFT, fill=Y)
        self.disAsmList.bind('<Double-Button-1>', self.AsmListDoubleClick)

        button = Button(frameBg, text='Call', command=self.Call, takefocus=0)
        button.grid(row=1, column=11, sticky=N+EW, rowspan=2, columnspan=1, padx=5)

        # Fill in row 2
        Label(frameBg, text='Y', bg=self.bg
              ).grid(row=2, column=0, sticky=E, columnspan=1, padx=5)
        yvalue = StringVar(top)
        yentry = Entry(frameBg, textvariable=yvalue, width=5)
        yentry.grid(row=2, column=1, sticky=W, columnspan=1, padx=5)
        l2value = StringVar(top)
        l2entry = Entry(frameBg, textvariable=l2value, width=12)
        l2entry.grid(row=2, column=2, sticky=W, columnspan=2, padx=5)
        m2value = StringVar(top)
        m2entry = Entry(frameBg, textvariable=m2value, width=3)
        m2entry.grid(row=2, column=4, sticky=W, columnspan=1, padx=5)

        # Fill in row 3
        Label(frameBg, text='X', bg=self.bg
              ).grid(row=3, column=0, sticky=E, columnspan=1, padx=5)
        xvalue = StringVar(top)
        xentry = Entry(frameBg, textvariable=xvalue, width=5)
        xentry.grid(row=3, column=1, sticky=W, columnspan=1, padx=5)
        l3value = StringVar(top)
        l3entry = Entry(frameBg, textvariable=l3value, width=12)
        l3entry.grid(row=3, column=2, sticky=W, columnspan=2, padx=5)
        m3value = StringVar(top)
        m3entry = Entry(frameBg, textvariable=m3value, width=3)
        m3entry.grid(row=3, column=4, sticky=W, columnspan=1, padx=5)
        button = Button(frameBg, text='Step', command=self.Step, takefocus=0)
        button.grid(row=3, column=11, sticky=N+EW, rowspan=2, columnspan=1, padx=5)

        # Fill in row 4
        Label(frameBg, text='A:B', bg=self.bg
              ).grid(row=4, column=0, sticky=E, columnspan=1, padx=5)

        frameDEntry = Frame(frameBg, relief=SUNKEN, borderwidth=2)
        frameDEntry.grid(row=4, column=1, sticky=W, rowspan=1, columnspan=1, padx=5)
        avalue = StringVar(top)
        aentry = Entry(frameDEntry, textvariable=avalue, width=2, relief=FLAT, borderwidth=0)
        aentry.pack(side=LEFT)
        Label(frameDEntry, text=':', bg = "white", padx=0, pady=0).pack(side=LEFT, fill=Y)
        bvalue = StringVar(top)
        bentry = Entry(frameDEntry, textvariable=bvalue, width=2, relief=FLAT, borderwidth=0)
        bentry.pack(side=LEFT)

        l4value = StringVar(top)
        l4entry = Entry(frameBg, textvariable=l4value, width=12)
        l4entry.grid(row=4, column=2, sticky=W, columnspan=2, padx=5)
        m4value = StringVar(top)
        m4entry = Entry(frameBg, textvariable=m4value, width=5)
        m4entry.grid(row=4, column=4, sticky=W, columnspan=1, padx=5)

        # Fill in row 5
        Label(frameBg, text='CC', bg=self.bg
              ).grid(row=5, column=0, sticky=E, columnspan=1, padx=5)
        ccvalue = StringVar(top)
        ccentry = Entry(frameBg, textvariable=ccvalue, width=3)
        ccentry.grid(row=5, column=1, sticky=W, columnspan=1, padx=5)
        l5value = StringVar(top)
        self.l5entry = Entry(frameBg, textvariable=l5value, width=12)
        self.l5entry.grid(row=5, column=2, sticky=W, columnspan=2, padx=5)
        m5value = StringVar(top)
        m5entry = Entry(frameBg, textvariable=m5value, width=5)
        m5entry.grid(row=5, column=4, sticky=W, columnspan=1, padx=5)
        button = Button(frameBg, text='Proceed', command=self.Proceed, takefocus=0)
        button.grid(row=5, column=11, sticky=N+EW, rowspan=2, columnspan=1, padx=5)

        # Fill in row 6
        Label(frameBg, text='SP', bg=self.bg
              ).grid(row=6, column=0, sticky=E, columnspan=1, padx=5)
        spvalue = StringVar(top)
        spentry = Entry(frameBg, textvariable=spvalue, width=5)
        spentry.grid(row=6, column=1, sticky=W, columnspan=1, padx=5)
        l6value = StringVar(top)
        l6entry = Entry(frameBg, textvariable=l6value, width=12)
        l6entry.grid(row=6, column=2, sticky=W, columnspan=2, padx=5)
        m6value = StringVar(top)
        m6entry = Entry(frameBg, textvariable=m6value, width=5)
        m6entry.grid(row=6, column=4, sticky=W, columnspan=1, padx=5)

        # Make lists of entry widgets for <Return> event handling
        # Register entry lists
        self.reglist = [pcentry, yentry, xentry, aentry,
                         bentry, ccentry, spentry]

        for entry in self.reglist:
            entry.bind('<Return>', self.ChangeReg) # Change any register
            entry.bind('<FocusOut>', self.ChangeReg)
            entry.bind('<FocusIn>', self.ReFocus)

        self.regentryl = [{'pyvar':pcvalue, 'regsize':2, 'regcmd':'RM P', 'replylen':56, 'previous':''},
                          {'pyvar':yvalue, 'regsize':2, 'regcmd':'RM Y', 'replylen':56, 'previous':''},
                          {'pyvar':xvalue, 'regsize':2, 'regcmd':'RM X', 'replylen':56, 'previous':''},
                          {'pyvar':avalue, 'regsize':1, 'regcmd':'RM A', 'replylen':54, 'previous':''},
                          {'pyvar':bvalue, 'regsize':1, 'regcmd':'RM B', 'replylen':54, 'previous':''},
                          {'pyvar':ccvalue, 'regsize':1, 'regcmd':'RM C', 'replylen':54, 'previous':''},
                          {'pyvar':spvalue, 'regsize':2, 'regcmd':'RM S', 'replylen':56, 'previous':''}]

        # Address entry lists
        self.addrlist = [l1entry, l2entry, l3entry, l4entry, self.l5entry, l6entry]
        for entry in self.addrlist:
            entry.bind('<Return>', self.ChangeAddr) # Display new memory location
            entry.bind('<FocusOut>', self.ChangeAddr)
            entry.bind('<FocusIn>', self.ReFocus)

        self.addentryl = [{'pyvar':l1value, 'previous':''},
                          {'pyvar':l2value, 'previous':''},
                          {'pyvar':l3value, 'previous':''},
                          {'pyvar':l4value, 'previous':''},
                          {'pyvar':l5value, 'previous':''},
                          {'pyvar':l6value, 'previous':''}]

        # Memory Entry and Value lists
        self.memlist = [m1entry, m2entry, m3entry, m4entry, m5entry, m6entry]
        for entry in self.memlist:
            entry.bind('<Return>', self.ChangeMem) # Change the contents of memory
            entry.bind('<FocusOut>', self.ChangeMem)
            entry.bind('<FocusIn>', self.ReFocus)

        self.mementryl = [{'pyvar':m1value, 'memsize':1, 'previous':''},
                          {'pyvar':m2value, 'memsize':1, 'previous':''},
                          {'pyvar':m3value, 'memsize':1, 'previous':''},
                          {'pyvar':m4value, 'memsize':2, 'previous':''},
                          {'pyvar':m5value, 'memsize':2, 'previous':''},
                          {'pyvar':m6value, 'memsize':2, 'previous':''}]

        # Fill in row 7
        Label(frameBg, text = 'Break Points', bg=self.bg
              ).grid(row=7, column=0, sticky=S+EW, columnspan=5, pady=5)
        button = Button(frameBg, text='Update', takefocus=0)
        button.grid(row=7, column=5, sticky=EW, columnspan=1, padx=5)
        button.bind('<ButtonRelease-1>', self.Update)
        button = Button(frameBg, text='Memory Dump', command=self.MemoryDump, takefocus=0)
        button.grid(row=7, column=6, sticky=EW, columnspan=1, padx=5)
        button = Button(frameBg, text='Block Fill', command=self.BlockFill, takefocus=0)
        button.grid(row=7, column=7, sticky=EW, columnspan=1, padx=5)
        button = Button(frameBg, text='Bulk Erase', command=self.BulkErase, takefocus=0)
        button.grid(row=7, column=8, sticky=EW, columnspan=1, padx=5)
        button = Button(frameBg, text='Dis-Assemble', command=self.DisAssemble, takefocus=0)
        button.grid(row=7, column=9, sticky=EW, columnspan=1, padx=5)
        button = Button(frameBg, text='Go', command=self.Go, takefocus=0)
        button.grid(row=7, column=11, sticky=N+EW, rowspan=2, columnspan=1, padx=5)

        # Fill in row 8
        frameBrEntry = Frame(frameBg, bg=self.bg)
        frameBrEntry.grid(row=8, column=0, sticky=NE+SW, rowspan=2, columnspan=5, padx=0)
        Label(frameBrEntry, text='BP1', bg=self.bg
              ).grid(row=0, column=0, sticky=E, columnspan=1, padx=0)
        bp1value = StringVar(top)
        bp1entry = Entry(frameBrEntry, textvariable=bp1value, width=12)
        bp1entry.grid(row=0, column=1, sticky=SW, columnspan=1, padx=5)
        Label(frameBrEntry, text='BP2', bg=self.bg
              ).grid(row=0, column=2, sticky=E, columnspan=1, padx=0)
        bp2value = StringVar(top)
        bp2entry = Entry(frameBrEntry, textvariable=bp2value, width=12)
        bp2entry.grid(row=0, column=3, sticky=SW, columnspan=1, padx=5)

        button = Button(frameBg, text='Clear', command=self.ClearBP, takefocus=0)
        button.grid(row=8, column=5, sticky=EW, rowspan=2, columnspan=1, padx=5)
        self.s19value = StringVar(top)
        self.s19entry = s19entry = Entry(frameBg, textvariable=self.s19value, width=20, takefocus=0)
        s19entry.grid(row=8, column=6, sticky=EW, rowspan=2, columnspan=4, padx=5)
        button = Button(frameBg, text='DnLoad', command=self.Load, takefocus=0)
        button.grid(row=8, column=10, sticky=EW, rowspan=1, columnspan=1, padx=5)
        button = Button(frameBg, text='Browse', command=self.Browse, takefocus=0)
        button.grid(row=8, column=11, sticky=EW, rowspan=1, columnspan=1, padx=5, pady=16)

        # Fill in row 9
        Label(frameBrEntry, text='BP3', bg=self.bg
              ).grid(row=1, column=0, sticky=E, columnspan=1, padx=0)
        bp3value = StringVar(top)
        bp3entry = Entry(frameBrEntry, textvariable=bp3value, width=12)
        bp3entry.grid(row=1, column=1, sticky=NW, columnspan=2, padx=5)
        Label(frameBrEntry, text='BP4', bg=self.bg
              ).grid(row=1, column=2, sticky=E, columnspan=1, padx=0)
        bp4value = StringVar(top)
        bp4entry = Entry(frameBrEntry, textvariable=bp4value, width=12)
        bp4entry.grid(row=1, column=3, sticky=NW, columnspan=2, padx=5)

        # Breakpoint Entry and Value lists
        self.bplist = [bp1entry, bp2entry, bp3entry, bp4entry]

        for entry in self.bplist:
            entry.bind('<Return>', self.ChangeBP) # Change a breakpoint
            entry.bind('<Tab>', self.ChangeBP)
            entry.bind('<FocusIn>', self.ReFocus)

        self.brkentryl = [{'pyvar':bp1value, 'previous':'', 'bp':''},
                          {'pyvar':bp2value, 'previous':'', 'bp':''},
                          {'pyvar':bp3value, 'previous':'', 'bp':''},
                          {'pyvar':bp4value, 'previous':'', 'bp':''}]


        # Fill in row 10
        Frame(frameBg, borderwidth=1, relief=SUNKEN,
              height=2, bg=self.bg).grid(row=10, column=0, sticky=EW,
                                         columnspan=12, padx=0, pady=0)  # Draw the separator

        # Fill in row 11
        self.buttonGUI = button = Button(frameBg, text='GUI Only',
                                         command=self.IGToggle, takefocus=0)
        button.grid(row=11, column=0, sticky=N, columnspan=3, padx=5, pady=3)
        self.usescivalue = IntVar(top)
        usescicb = Checkbutton(frameBg, variable=self.usescivalue, text='Use SCI', bg=self.bg)
        usescicb.grid(row=11, column=3, sticky=SE, columnspan=1, padx=5, pady=0)
        self.termlable = Label(frameBg, text='HC11 SCI - Terminal', bg=self.bg)
        self.termlable.grid(row=11, column=5, sticky=S+EW, columnspan=3, padx=1)
        Label(frameBg, text='Baud Rate', bg=self.bg
              ).grid(row=11, column=10, sticky=SE, columnspan=1, padx=5)
        self.baudrateValue = IntVar(top)
        self.baudrateValue.set(self.baudrate)
        brentry = Entry(frameBg, textvariable=self.baudrateValue, width=6, takefocus=0)
        brentry.grid(row=11, column=11, sticky=SW, columnspan=1, padx=5)

        brentry.bind('<Return>', self.ChangeBaud)
        brentry.bind('<Tab>', self.ChangeBaud)
        brentry.bind('<FocusIn>', self.ReFocus)

        self.frameText = frameText = Frame(top, relief=FLAT)
        self.scrollbarView = Scrollbar(frameText, orient=VERTICAL,
                                       takefocus=FALSE, highlightthickness=0)
        self.textView = textView = Text(frameText, wrap=WORD, highlightthickness=0,
                                        bd=2, font=('Courier', 10))
        self.scrollbarView.config(command=textView.yview)
        textView.config(yscrollcommand=self.scrollbarView.set)
        self.scrollbarView.pack(side=RIGHT, fill=Y)
        textView.pack(side=LEFT, expand=TRUE, fill=BOTH)
        frameText.grid(row=1, column=0, sticky=EW)

        textView.bind('<Control-KeyPress>', self.OnTermControlKeyPress)
        textView.bind('<KeyPress>', self.OnTermKeyPress)
        textView.bind('<Return>', self.OnTermReturnKey)
        textView.bind('<Tab>', self.OnTermTabKey)

###################################################################

    def notresponding(self):
        message = "The targer is not responding"
        m = tkMessageBox.Message(
            title = "Communications Timeout",
            message = message,
            icon = tkMessageBox.WARNING,
            type = tkMessageBox.OK,
            master = self.top)
        return m.show()

    # Act as output file
    def write(self, s, tags=(), mark="insert"):
        self.textView.insert(mark, s, tags)
        self.textView.see(mark)
        self.textView.update()

    def writelines(self, l):
        map(self.write, l)

    def ReFocus(self, event):
        event.widget.select_range(0, END)

### NOT USED ### Save as an example of an event handler that
    def UnFocus(self, event):   # uses this method for class bindings
        regentryl = self.regentryl
        addentryl = self.addentryl
        mementryl = self.mementryl
        thisentry = event.widget
        try:
            teindex = self.reglist.index(thisentry)
            oldval = regentryl[teindex]['previous']
            newval = regentryl[teindex]['pyvar'].get()
            if oldval != newval:
                return self.ChangeReg(thisentry=thisentry)
            return
        except ValueError:
            pass
        try:
            teindex = self.addrlist.index(thisentry)
            oldval = addentryl[teindex]['previous']
            newval = addentryl[teindex]['pyvar'].get()
            if oldval != newval:
                return self.ChangeAddr(thisentry=thisentry)
            return
        except ValueError:
            pass
        try:
            teindex = self.memlist.index(thisentry)
            oldval = mementryl[teindex]['previous']
            newval = mementryl[teindex]['pyvar'].get()
            if oldval != newval:
                return self.ChangeMem(thisentry=thisentry)
            return
        except ValueError:
            pass

    def OnTermControlKeyPress(self, event):
        keysym = event.keysym
        key = ''
        if keysym == 'x':
            key = chr(0x18)
        elif keysym == 'w':
            key = chr(0x17)
        elif keysym == 'd':
            key = chr(0x04)
        elif keysym == 'c':
            key = chr(0x03)
        elif keysym == 'b':
            key = chr(0x02)
        elif keysym == 'a':
            key = chr(0x01)
        if key:
            self.SendChars(key)
        if keysym < 'c':
            return 'break'

    def OnTermTabKey(self, event):
        self.reglist[0].focus_set()
        return 'break'

    def OnTermReturnKey(self, event):
        self.OnTermKeyPress(event)
        return 'break'

    def OnTermKeyPress(self, event):
        # move insertion point to the end
        self.textView.mark_set("insert", END)
        keysym = event.keysym
        if keysym not in notkeys:
#            print event.char, keysym, event.keycode,
            key = event.char
            if keysym == 'Delete':
#                print event.keycode
                key = chr(0x7F)
            self.SendChars(key)
        if keysym != 'BackSpace':
            return 'break'

    def AsmListDoubleClick(self, event):
        asmlist = event.widget
        asmline = asmlist.get(asmlist.curselection())
        self.regentryl[0]['pyvar'].set(asmline[:4])
        self.ChangeReg(thisentry=self.reglist[0])

    def Browse(self, event=None):
        srecname = tkFileDialog.Open(filetypes = [("S records", "*.s19", "TEXT")]).show()
        self.s19value.set(srecname)
        return 'break'

    def IGToggle(self, event=None):
        if not self.gui_only:
            self.top.resizable(height=TRUE, width=TRUE)    # unlock the size
            self.frameText.grid_remove()
            self.top.resizable(height=FALSE, width=FALSE)    # then lock the size
            self.buttonGUI.config(text="Interactive")
            self.termlable.config(text="")
        else:
            self.top.resizable(height=TRUE, width=TRUE)    # unlock the size
            self.frameText.grid()
            self.top.resizable(height=FALSE, width=FALSE)    # then lock the size
            self.buttonGUI.config(text="GUI Only")
            self.termlable.config(text="HC11 SCI - Terminal")
        self.gui_only = not self.gui_only
        self.top.wm_deiconify()

    def ChangeBaud(self, event):
        oldfw = self.top.focused_widget
        self.top.focused_widget = thisentry = event.widget
        baud = self.baudrateValue.get()
        if baud == self.baudrate:
            return
        if not baud in (9600, 4800, 2400, 1200, 600, 300):
            thisentry.select_range(0, END)
            self.top.bell()
            return
        multi = 9600//baud
        divis = -1
        while multi:
            multi = multi >> 1
            divis += 1
        cmd = 'MM 102B\r'
        arg = '3' + str(divis)          # Write target memory
        reply = self.SendCommand(cmd, replylen=17)
        if not reply:
            return
        self.Converse(arg)
        self.SendChars('\r')
        self.baudrate  = baud
        quetime = 20.0/baud + .01
        if quetime > self.quetime:
            self.quetime = quetime      # Wait as long as possible
        reply, t = self.PollQueue(expect=10)
        if quetime != self.quetime:
            self.quetime = quetime
        self.tty.setbaudrate(baud)
        self.top.focused_widget = oldfw

    def MemoryDump(self, event=None):
#      '0000 AA 55 38 72 86 AA 97 00 C6 55 D7 01 3D DD 02 39  U8r     U  =  9        '
#       78 characters X 9 lines = 702 + newline and prompt = 704
        addr = BCDbaseDialog.askhex("Dump Address", "Starting Address: 0 - FFFF",
                                    parent=self.textView, minvalue=0, maxvalue=0xffff)
        if addr:
            self.SendCommand('MD %s\r' %addr)

    def BlockFill(self, event=None):
        block = BCDbaseDialog.askblockvalue('Block Fill',
                                            'Use hex: Start < End < $8000',
                                            parent=self.disAsmList)
        if block and block[0]:
            self.SendCommand('BF ' + ' '.join(block) + '\r', replylen=16)

    def BulkErase(self, event=None):
        self.SendCommand('BULK\r', replylen=4)

    def DisAssemble(self, event=None):
        addr = BCDbaseDialog.askhex("Dis-Assembler", "Starting Address: 0 - FFFF",
                                    parent=self.textView, minvalue=0, maxvalue=0xffff)
        if addr:
            self.disAsmList.delete(0, END)
            self.asmAddrList = []
            count = askinteger("Dis-Assembler", "Number of Lines: 0 - 255",
                               parent=self.textView,
                               initialvalue=1, minvalue=0, maxvalue=0xff)
            reply = self.SendCommand('ASM %s\r' %addr, replylen=35)
            if not reply:
                return
            try:
                reply = reply.split('\n')[-1]
            except IndexError:
                return
            for i in range(count):
                self.disAsmList.insert(i, reply[:-1].rstrip())
                self.asmAddrList.append(reply[:4])
                self.root.update()
                reply = self.Converse('\r', replylen=35).split('\n')[-1]
                

    def Call(self, event=None):
        capture = not self.usescivalue.get()
        replen = 65 * capture
        reply = self.SendCommand('CALL\r', replylen=replen)
        if capture:
            good, registers = self.SplitRegisters('Call', reply)
            if good:
                self.Update(registers=registers)

    def Step(self, event=None):
#       t

#       BRCL $00,X $01 $E00A  P-E00A Y-2222 X-100A A-38 B-72 C-90 S-0047 '
#       >
        reply = self.SendCommand('T\r', replylen=70)
        good, registers = self.SplitRegisters('Trace', reply)
        if good:
            self.Update(registers=registers)

    def Proceed(self, event=None):
#       p

#       P-E365 Y-0005 X-00A6 A-0A B-09 C-90 S-003E '
#       >
        reply = self.SendCommand('P\r', replylen=48)
        good, registers = self.SplitRegisters('Proceed', reply)
        if good:
            self.Update(registers=registers)

    def Go(self, event=None):
        reply = self.SendCommand('G\r', replylen=69)
        good, registers = self.SplitRegisters('Go', reply)
        if good:
            self.Update(registers=registers)

    def Load(self, event=None):
        fname = self.s19value.get()
        if fname:
            self.SendCommand('LOAD T\r')
            try:
                tmpfile = open(fname)
                self.top.config(cursor="watch")
                for line in tmpfile:
                    self.root.update()
                    for char in line:
                        self.SendChars(char)
            except IOError:
                self.top.bell()
                self.s19entry.select_range(0, END)
                return
            tmpfile.close()
            reply = self.Converse('', replylen=18)     # The promt will arrive
            if not reply:
                return
            rc = len(reply)
            if not rc:
                return
            if rc > 7:
                self.top.bell()
                if rc == 12:
                    self.Converse('\r', replylen=9)
                    print >> self, 'Download Error: Read Only Memory'
                elif rc > 12:
                    print >> self, 'Download Error: bad checksum'
                else:
                    print >> self, 'Download Error: the target isn\'t ready'
                self.top.config(cursor="")
                return
            cmd = 'BR -\r'  # Clear breakpoints
            self.SendCommand(cmd, replylen=29)
            for entrydict in self.brkentryl:    # Clear all entry widgets
                entrydict['pyvar'].set('')
                entrydict['previous'] = ''
                entrydict['bp'] = ''
            for entrydict in self.addentryl:
                entrydict['pyvar'].set('')
                entrydict['previous'] = ''
            for entrydict in self.mementryl:
                entrydict['pyvar'].set('')
                entrydict['previous'] = ''
            for entrydict in self.regentryl:
                entrydict['pyvar'].set('')
                entrydict['previous'] = ''
                self.disAsmList.delete(0, END)
            self.labelxref = {}          # Dictionary of labels and their values
            self.constxref = {}          # Dictionary of constants and their values
            self.inversexref = {}         # Inverse dictionary of labels and their values
            self.asmAddrList = []   # List of address in the loaded context
            fname = fname[:-4] + '.lst'
            tmpfile = open(fname)
            for i, line in enumerate(tmpfile):
                self.disAsmList.insert(i, line[6:-1])   # Instert line into list box
                self.asmAddrList.append(line[6:10])     # Append to list for address lookup
                self.root.update()
            tmpfile.close()
            fname = fname[:-4] + '.xrf'
            tmpfile = open(fname)
            tmpfile.readline()      # Skip Labels header lines of the file
            tmpfile.readline()      # .readline() can be used before .next()
            tmpfile.readline()
            for line in tmpfile:    # for loops call .next() repeatedly
                if line[0] == ' ':
                    continue
                lablinelist = line.split()
                try:
                    self.labelxref[lablinelist[0]] = lablinelist[1][1:]
                    self.inversexref[lablinelist[1][1:]] = lablinelist[0]
                except IndexError:
                    break
                self.root.update()
            tmpfile.next()          # Skip Constants header lines of the file
            tmpfile.next()          # .readline() can NOT be used after .next()
            tmpfile.next()          # because of look-ahead buffering
            for line in tmpfile:
                if line[0] == ' ':
                    continue
                lablinelist = line.split()
                try:
                    self.constxref[lablinelist[0]] = lablinelist[1][1:]
                    self.inversexref[lablinelist[1][1:]] = lablinelist[0]
                except IndexError:
                    break
                self.root.update()
            tmpfile.close()
            self.top.config(cursor="")
#            self.top.bell()
        else:
            self.Browse()

    def ChangeBP(self, event):
#       br 0001
#
#
#       0001 0000 0000 0000 
#       >
        thisentry = event.widget
        teindex = self.bplist.index(thisentry)
        tedict = self.brkentryl[teindex]
        chars = tedict['pyvar'].get()
        if chars != tedict['previous']:
            oldbp = tedict['bp']
            if not chars:
                if oldbp:
                    self.RemoveBP(oldbp, teindex)
            else:
                try:
                    a = self.labelxref[chars]
                    for i, entrydict in enumerate(self.brkentryl):
                        if a == entrydict['bp']:
                            if i != teindex:
                                raise KeyError
                    tedict['previous'] = chars
                    chars = a
                except KeyError:
                    try:
                        a = int(chars, 16)
                        if a > 0x7FFF:
                            raise ValueError
                        chars = hex(a)[2:].upper().zfill(4)
                        for i, entrydict in enumerate(self.brkentryl):
                            if chars == entrydict['bp']:
                                if i != teindex:
                                    raise ValueError
                                entrydict['pyvar'].set(entrydict['previous'])
                                return
                        try:
                            a = self.inversexref[chars]
                        except KeyError:
                            a = chars
                        tedict['pyvar'].set(a)
                        tedict['previous'] = a
                    except ValueError:
                        thisentry.bell()
                        thisentry.focus_set()
                        thisentry.select_range(0, END)
                        return "break"
                if chars != oldbp:
                    if oldbp:
                        self.RemoveBP(oldbp, teindex)
                    cmd = 'BR %s\r' %chars
                    reply = self.SendCommand(cmd, replylen=31)
                    if not reply:
                        return
                    tedict['bp'] = chars
        teindex = self.NextEntry(teindex, 3, event.state & 1)
        self.bplist[teindex].focus_set()
        self.bplist[teindex].select_range(0, END)

    def RemoveBP(self, oldbp, index):
        cmd = 'BR -%s\r' %oldbp
        self.SendCommand(cmd, replylen=33)
        self.brkentryl[index]['bp'] = ''

    def ClearBP(self, event=None):
        bpelist = self.brkentryl
        reply = self.SendCommand('BR -\r', replylen=33)
        if reply:
            for i in (0, 1, 2, 3):
                bpelist[i]['bp'] = ''
                bpelist[i]['previous'] = ''
                bpelist[i]['pyvar'].set('')

    def ChangeReg(self, event=None, thisentry=None):
        if event:
            thisentry = event.widget
        teindex = self.reglist.index(thisentry)
        tedict = self.regentryl[teindex]
        chars = tedict['pyvar'].get()
        if chars != tedict['previous']:
            if chars:
                size = tedict['regsize']
                try:
                    a = int(chars, 16)
                    if a > 256**(tedict['regsize'])-1:
                        raise ValueError
                    chars = hex(a)[2:].upper().zfill(size * 2)
                    tedict['pyvar'].set(chars)
                    tedict['previous'] = chars
                except ValueError:
                    thisentry.bell()
                    thisentry.select_range(0, END)
                    return "break"
                if teindex == 0:
                    self.SelectNextAsmLine(chars)
                cmd = tedict['regcmd'] + '\r'
                replen = tedict['replylen']
                reply = self.SendCommand(cmd, replylen=replen)
                if not reply:
                    return
                self.Converse(chars + '\r', replylen=8)
        if event and event.type != '10':
            teindex = self.NextEntry(teindex, 6, event.state & 1)
            self.reglist[teindex].focus_set()
            self.reglist[teindex].select_range(0, END)

    def ChangeMem(self, event=None, thisentry=None):
#       'mm                     3 * 2
#                               1
#       0005 06 '' FF ''        12
#                               1
#                               1
#       rom-                    5
#       >'                      1
        if event:
            thisentry = event.widget
        teindex = self.memlist.index(thisentry)
        tedict = self.mementryl[teindex]
        chars = tedict['pyvar'].get()
        if chars != tedict['previous']:
            try:
                addr = tedict['address']
            except KeyError:    # This happens when address entry has not been varified
                if not chars:
                    return
                thisentry.bell()
                self.addrlist[teindex].focus_set()
                self.addrlist[teindex].select_range(0, END)
                return
            try:
                a = int(chars, 16)
                size = tedict['memsize']
                if a > 256**size -1:
                    raise ValueError
                chars = hex(a)[2:].upper().zfill(size * 2)
                tedict['pyvar'].set(chars)
                tedict['previous'] = chars
            except ValueError:
                if not chars:
                    tedict['pyvar'].set(tedict['previous'])
                thisentry.bell()
                thisentry.focus_set()
                thisentry.select_range(0, END)
                return "break"
            cmd = 'MM %s\r' %addr            # Write target memory
            self.SendCommand(cmd, 17)
            reply = self.Converse(chars[:2] + ' ', replylen=8)
            if not reply:
                return
            rc = len(reply)
            if rc > 6:
                thisentry.bell()
                tedict['pyvar'].set('ROM')
                tedict['previous'] = 'ROM'
                thisentry.select_range(0, END)
                self.Converse('\r', replylen=3)
                return
            if size == 2:
                self.Converse(chars[2:4])
            reply = self.Converse('\r', replylen=8)
            rc = len(reply)
            if rc > 6:
                thisentry.bell()
                tedict['pyvar'].set(chars[:2] + 'ROM')
                tedict['previous'] = chars[:2] + 'ROM'
                thisentry.select_range(0, END)
                return
        if event and event.type != '10':
            teindex = self.NextEntry(teindex, 5, event.state & 1)
            self.memlist[teindex].focus_set()
            self.memlist[teindex].select_range(0, END)

    def ChangeAddr(self, event=None, thisentry=None):
#       'mm
#
#       0005 06 '' FF ''
#
#       >'
        if event:
            thisentry = event.widget
        teindex = self.addrlist.index(thisentry)
        tedict = self.addentryl[teindex]
        memdict = self.mementryl[teindex]
        chars = tedict['pyvar'].get()
        if chars != tedict['previous']:
            if chars:
                try:
                    chars = self.constxref[chars]
                except KeyError:
                    try:
                        a = int(chars, 16)
                        if a > 0xFFFF:
                            raise ValueError
                        chars = hex(a)[2:].upper().zfill(4)
                        try:
                            a = self.inversexref[chars]
                        except KeyError:
                            a = chars
                        tedict['pyvar'].set(a)
                        tedict['previous'] = a
                    except ValueError:
                        memdict['pyvar'].set('')
                        memdict['previous'] = ''
                        memdict.pop('address', 0)    # We don't care if it didn't exist.
                        thisentry.bell()
                        thisentry.focus_set()
                        thisentry.select_range(0, END)
                        return "break"
                memdict['address'] = chars
                size = memdict['memsize']           # Read target memory
                reply = self.SendCommand('MM %s\r' %chars, replylen=17)
                if not reply:
                    return
                value = reply[-3:-1]
                if size == 2:
                    reply = self.Converse(' ', replylen=4)
                    value += reply[-3:-1]
                memdict['pyvar'].set(value)
                memdict['previous'] = value
                self.Converse('\r', replylen=4)
            else:
                memdict['pyvar'].set('')
                memdict['previous'] = ''
                memdict.pop('address', 0)    # We don't care if it didn't exist.
        if event and event.type != '10':
            teindex = self.NextEntry(teindex, 5, event.state & 1)
            self.addrlist[teindex].focus_set()
            self.addrlist[teindex].select_range(0, END)

    def Update(self, event=None, registers=[]):
#   send 4, then listen for
#       RM                      (3)
#       P-FFFF Y-FFFF X-FFFF A-FF B-FF C-D0 S-0047 (43)
#       P-FFFF                  (7)
#                               (1)
#       >                       (1)
        if event:
            cmd = 'RM\r'
            reply = self.SendCommand(cmd, replylen=58)
            if not reply:
                return
            good, registers = self.SplitRegisters('RM', reply)
            self.Converse('\r', replylen=4)
            if not good:
                return
        for i, regdict in enumerate(self.regentryl):
            chars = registers[i][2:]
            if i == 0:
                self.SelectNextAsmLine(chars)
            regdict['pyvar'].set(chars)
            regdict['previous'] = chars
#       'MM                     (3)
#                               (1)
#                               (1)
#       0005 06 '' FF ''        (12)
        for memdict in self.mementryl:
            try:
                addr = memdict['address']
            except KeyError:
                continue    # This happens when address entry has not been varified
            else:
                size = memdict['memsize']    # Read target memory
                reply = self.SendCommand('MM %s\r' %addr, replylen=17)
                value = reply[-3:-1]
                if size == 2:
                    reply = self.Converse(' ', replylen=4)
                    value += reply[-3:-1]
                memdict['pyvar'].set(value)
                memdict['previous'] = value
                self.Converse('\r', replylen=4)

    def SplitRegisters(self, command, reply):
        offset = 0
        lineindex = -2
        if command == 'Trace':
            offset = 21
        if command == 'RM':
            lineindex = 0
        goodanswer = False
        try:
            registers = reply.split('\n')[lineindex]
            registers = registers[offset:].split()
            if len(registers) != 7:
                raise IndexError
            goodanswer = True
        except IndexError:
            print >> self, 'Got a bad reply from %s:' % command
            registers = []
        return goodanswer, registers

    def NextEntry(self, teindex, lastentry, shift):
        if shift:  # Shift key is down
            if teindex == 0:
                teindex = lastentry
            else:
                teindex -= 1
        else:
            if teindex == lastentry:
                teindex = 0
            else:
                teindex += 1
        return teindex

    def SelectNextAsmLine(self, next):
        self.disAsmList.select_clear(ACTIVE)
        try:
            nextline = self.asmAddrList.index(next)
        except ValueError:
            pass
        else:
            self.disAsmList.select_set(nextline)    # Doesn't activate the line
            self.disAsmList.activate(nextline)  # Doesn't highlight the line
            self.disAsmList.see(nextline)

##########################################

    def SendChars(self, chars):
        self.tty.write(chars)       # Write to tty

    def SendCommand(self, chars, replylen=0):
        self.textView.mark_set("insert", END)
        self.Converse(chr(0x18), replylen=2)    # Terminate any pending command
        reply = self.Converse(chars, replylen=replylen)
        return reply

    def Converse(self, cmd, replylen=0, notify=False):
        count = 0
        replystring = ''
        if cmd:
            nchars = len(cmd)
            self.tty.write(cmd)       # Write to tty if any
            chars, timeout = self.PollQueue(nchars)
            if timeout:
                self.notresponding()
                self.top.wakeup()
                return ''
        if replylen:
            chars, timeout = self.PollQueue(replylen, wait_one=True)
            if timeout:
#                print chars + '\n', len(chars), 'chars, out of %d expected' %replylen
                if notify:
                    self.notresponding()
                    self.top.wakeup()
                    return ''
        return chars

    def PeriodicCall(self):
        if self.running:
            self.PCID = self.root.after(50, self.PeriodicCall)
            self.PollQue()    # Call abreviated PQ

    def PollQueue(self, expect, wait_one=False):
        count = 0
        outstring = ''
        if wait_one:
            endtime = time() + 0.100
            while not self.queue.qsize():
                if time() > endtime:
                    break
        try:
            while count < expect:
                timeout = True
                data = self.queue.get(timeout=self.quetime)
                timeout = False
                for char in data:
                    if ord(char) <= 12:
                        continue
                    if ord(char) == 13:
                        char = '\n'
                    elif ord(char) <= 0x19:
                        continue
                    outstring += char
                    count += 1
        finally:
            return outstring, timeout

##
##        if wait_one:
##            self.root.after_cancel(self.PCID)
##            while not self.queue.qsize():
##                self.root.update()
##            self.root.after(50, self.PeriodicCall)
##        


    def PollQue(self):
        outstring = ''
        while self.queue.qsize():
            data = self.queue.get(timeout=self.quetime)
            for char in data:
                if ord(char) <= 12:
                    continue
                if ord(char) == 13:
                    char = '\n'
                elif ord(char) <= 0x19:
                    continue
                outstring += char
        if outstring:
            self.write(outstring)

    def close(self, event=None):
        self.running = False
        self.root.after_cancel(self.PCID)
        self.tty.close()
        if self.flist:
            self.flist.debugger = None
        self.queue = None
        self.top.destroy()


# A basic synchronous COM port file-like object
class SerialTTY:
    def __init__(self, master, port, baud):
        self.master = master
        self.port = port
        self.running = master.running
        self.eventStop = CreateEvent(None, 0, 0, None)
        self.overlapped = OVERLAPPED()
        self.overlapped.hEvent = CreateEvent(None, 1, 0, None)
        self.handle = CreateFile(port,
                                 win32con.GENERIC_READ | win32con.GENERIC_WRITE,
                                 0, # exclusive access
                                 None, win32con.OPEN_EXISTING,
                                 win32con.FILE_ATTRIBUTE_NORMAL | win32con.FILE_FLAG_OVERLAPPED,
                                 None)
        SetCommMask(self.handle, EV_RXCHAR) # Tell the port we want a notification on each char.
        SetupComm(self.handle, 4096, 4096)  # Setup a 4k buffer
        # Remove anything that was there
        PurgeComm(self.handle, PURGE_TXABORT | PURGE_RXABORT | PURGE_TXCLEAR | PURGE_RXCLEAR )
        timeouts = 0xFFFFFF, 0, 1000, 0, 1000   # ReadIntervalTimeout
        SetCommTimeouts(self.handle, timeouts)  # ReadTotalTimeoutMultiplier, ReadTotalTimeoutConstant
        dcb = GetCommState( self.handle )       # WriteTotalTimeoutMultiplier, WriteTotalTimeoutConstant
        dcb.BaudRate = baud       # Setup the connection info.
        dcb.ByteSize = 8
        dcb.Parity = NOPARITY
        dcb.StopBits = ONESTOPBIT
        SetCommState(self.handle, dcb)
        print "Connected to %s at %s baud" % (port, dcb.BaudRate)

    def setbaudrate(self, baud):
        if baud > 9600 or baud % 300:
            print "Bad baud rate!"
            return
        PurgeComm(self.handle, PURGE_TXABORT | PURGE_RXABORT | PURGE_TXCLEAR | PURGE_RXCLEAR )
        dcb = GetCommState( self.handle )
        dcb.BaudRate = baud
        SetCommState(self.handle, dcb)
        print "Connected to %s at %s baud" % (self.port, dcb.BaudRate)

    def write(self, chars):
        try:
            WriteFile(self.handle, chars, self.overlapped)
            # Wait for the write to complete.
            WaitForSingleObject(self.overlapped.hEvent, INFINITE)
        finally:
            pass

    def read(self):
        overlapped = OVERLAPPED()
        overlapped.hEvent = CreateEvent(None, 1, 0, None)
        while self.running:
            rc, mask = WaitCommEvent(self.handle, overlapped)
            if rc == 0: # Character already ready!
                SetEvent(overlapped.hEvent)
            rc = WaitForMultipleObjects([overlapped.hEvent, self.eventStop], 0, INFINITE)
            if rc == WAIT_OBJECT_0:
                # Some input - read and print it
                flags, comstat = ClearCommError( self.handle )
                rc, data = ReadFile(self.handle, comstat.cbInQue, overlapped)
                WaitForSingleObject(overlapped.hEvent, INFINITE)
                self.master.queue.put(str(data))
            else:
                break   # Stop the thread!
            
    def close(self):
        SetEvent(self.eventStop)
        self.handle = CloseHandle(self.handle)

if __name__ == '__main__':
    # test the dialog
    import Tix
    root = Tix.Tk()
    dbug = Debugger(root)
    root.mainloop()
    root.destroy()
