from WindowList import ListedToplevel
from progressDialog import progressDialog
import tkFileDialog
import tkMessageBox
from Tkinter import *
from Queue import *
import threading
from time import time, sleep
from win32file import *     # The base COM port and file IO functions.
from win32event import *    # We use events and the WaitFor[Multiple]Objects functions.
import win32con             # constants.


notkeys = set(['Shift_L', 'Shift_R', 'Control_R', 'Control_L', 'Win_L', 'Win_R', 'Prior', 'Next',
               'Alt_L', 'Alt_R', 'Tab', 'Right', 'Left','Up', 'Down', 'Escape', 'Insert', 'Home',
               'F1', 'F2', 'F3', 'F4', 'F5', 'F6', 'F7', 'F8', 'F9', 'F10', 'L1', 'L2', 'End'])

class Debugger:
    """Dialog Window for BCD11IDE's debugger.

    """
    def __init__(self, root, menu=None, flist=None, srecord=''):
        self.root = root
        self.top = top = ListedToplevel(root, menu=menu)
        top.option_add("*Entry*font", "Courier 10", 20)
        top.configure(borderwidth=5)
        top.geometry("+%d+%d" % (root.winfo_rootx()+75,
                                  root.winfo_rooty()+100))
        if menu is None:
            root.withdraw()     # Here, we are running from __main__
        self.bg = "white"
        self.fg = "black"
        self.baudrate = baudrate = 9600     # Needed by baudrate value widget
        self.CreateWidgets(top)                     # Fit widgets into self
        top.resizable(height=FALSE, width=FALSE)    # then lock the size
        top.title('BCD11IDE Debugger')
        top.protocol("WM_DELETE_WINDOW", self.close)
        top.tkraise()
        top.wm_deiconify()

        self.s19value.set(srecord)
        top.focused_widget = self.s19entry
        self.top.focused_widget.focus_set()

        self.labelxref = {}          # Dictionary of labels and their values
        self.constxref = {}          # Dictionary of constants and their values

        self.flist = flist
        self.running = True
        self.queue = queue = Queue()            # Queue for inter-tread communication
        self.tty = SerialTTY(self, "COM1", baudrate)    # Pass que be reference?
        self.listener = threading.Thread(target=self.tty.read)
        self.listener.start()                   # Call tty.read in a backgriund thread!

        self.catchecho = self.replycount = 0    # Number of characters expected from target

        self.top.after_idle(self.Load)
        self.PeriodicCall()


    def CreateWidgets(self, top):
        self.gui_only = False
        self.frameMain = frameMain = Frame(top, borderwidth=2, relief=SUNKEN)
        frameMain.grid(row=0, column=0)
        frameBg = Frame(frameMain, bg = self.bg)  # Make the background frame
        frameBg.pack(expand=TRUE, fill=BOTH)
        # Put labels across row zero
        Label(frameBg, text = 'Registers', bg = self.bg
              ).grid(row=0, column=0, sticky=EW, columnspan=2, padx=5, pady=5)
        Label(frameBg, text='Addr / Label', bg = self.bg
              ).grid(row=0, column=2, sticky=W, columnspan=2, padx=5, pady=5)
        Label(frameBg, text='Value', bg = self.bg
              ).grid(row=0, column=4, sticky=W, columnspan=1, padx=5, pady=5)
        Label(frameBg, text='Address', bg = self.bg
              ).grid(row=0, column=5, sticky=W, columnspan=1, padx=5, pady=5)
        Label(frameBg, text=' Machine  Code ', bg = self.bg
              ).grid(row=0, column=6, sticky=W, columnspan=1, padx=5, pady=5)
        Label(frameBg, text='Label  Name ', bg = self.bg
              ).grid(row=0, column=7, sticky=W, columnspan=1, padx=5, pady=5)
        Label(frameBg, text='Mnemonic    Operands ', bg = self.bg
              ).grid(row=0, column=8, sticky=EW, columnspan=2, padx=5, pady=5)
        Label(frameBg, text='Command', bg = self.bg
              ).grid(row=0, column=11, sticky=W, columnspan=1, padx=5, pady=5)

        # Fill in row 1
        Label(frameBg, text='PC', bg = self.bg
              ).grid(row=1, column=0, sticky=E, columnspan=1, padx=5)
        pcvalue = StringVar(top)
        pcentry = Entry(frameBg, textvariable=pcvalue, width=5)
        pcentry.grid(row=1, column=1, sticky=W, columnspan=1, padx=5)
        l1value = StringVar(top)
        l1entry = Entry(frameBg, textvariable=l1value, width=12)
        l1entry.grid(row=1, column=2, sticky=W, columnspan=2, padx=5)
        m1value = StringVar(top)
        m1entry = Entry(frameBg, textvariable=m1value, width=3)
        m1entry.grid(row=1, column=4, sticky=W, columnspan=1, padx=5)

        frameDisAsm = Frame(frameBg)
        frameDisAsm.grid(row=1, column=5, sticky=N+SW, rowspan=6, columnspan=6, padx=5)
        self.disAsmList = Listbox(frameDisAsm, width=60, height=8, takefocus=FALSE,
                                  font=('Courier', 8), exportselection=FALSE)
        self.dalScroll = Scrollbar(frameDisAsm, command=self.disAsmList.yview)
        self.disAsmList.config(yscrollcommand=self.dalScroll.set)
        self.disAsmList.pack(side=LEFT, expand=TRUE, fill=BOTH)
        self.dalScroll.pack(side=LEFT, fill=Y)
        self.disAsmList.bind('<Double-Button-1>', self.AsmListDoubleClick)

        button = Button(frameBg, text='Call', command=self.Call, takefocus=0)
        button.grid(row=1, column=11, sticky=N+EW, rowspan=2, columnspan=1, padx=5)

        # Fill in row 2
        Label(frameBg, text='Y', bg = self.bg
              ).grid(row=2, column=0, sticky=E, columnspan=1, padx=5)
        yvalue = StringVar(top)
        yentry = Entry(frameBg, textvariable=yvalue, width=5)
        yentry.grid(row=2, column=1, sticky=W, columnspan=1, padx=5)
        l2value = StringVar(top)
        l2entry = Entry(frameBg, textvariable=l2value, width=12)
        l2entry.grid(row=2, column=2, sticky=W, columnspan=2, padx=5)
        m2value = StringVar(top)
        m2entry = Entry(frameBg, textvariable=m2value, width=3)
        m2entry.grid(row=2, column=4, sticky=W, columnspan=1, padx=5)

        # Fill in row 3
        Label(frameBg, text='X', bg = self.bg
              ).grid(row=3, column=0, sticky=E, columnspan=1, padx=5)
        xvalue = StringVar(top)
        self.xentry = Entry(frameBg, textvariable=xvalue, width=5)
        self.xentry.grid(row=3, column=1, sticky=W, columnspan=1, padx=5)
        l3value = StringVar(top)
        l3entry = Entry(frameBg, textvariable=l3value, width=12)
        l3entry.grid(row=3, column=2, sticky=W, columnspan=2, padx=5)
        m3value = StringVar(top)
        m3entry = Entry(frameBg, textvariable=m3value, width=3)
        m3entry.grid(row=3, column=4, sticky=W, columnspan=1, padx=5)

        button = Button(frameBg, text='Step', command=self.Step, takefocus=0)
        button.grid(row=3, column=11, sticky=N+EW, rowspan=2, columnspan=1, padx=5)

        # Fill in row 4
        Label(frameBg, text='A:B', bg = self.bg
              ).grid(row=4, column=0, sticky=E, columnspan=1, padx=5)
        frameDEntry = Frame(frameBg, relief=SUNKEN, borderwidth=2)

        frameDEntry.grid(row=4, column=1, sticky=W, rowspan=1, columnspan=1, padx=5)
        avalue = StringVar(top)
        self.aentry = Entry(frameDEntry, textvariable=avalue, width=2, relief=FLAT, borderwidth=0)
        self.aentry.pack(side=LEFT)

        Label(frameDEntry, text=':', bg = self.bg, padx=0, pady=0).pack(side=LEFT, fill=Y)

        bvalue = StringVar(top)
        self.bentry = Entry(frameDEntry, textvariable=bvalue, width=2, relief=FLAT, borderwidth=0)
        self.bentry.pack(side=LEFT)

        l4value = StringVar(top)
        l4entry = Entry(frameBg, textvariable=l4value, width=12)
        l4entry.grid(row=4, column=2, sticky=W, columnspan=2, padx=5)
        m4value = StringVar(top)
        m4entry = Entry(frameBg, textvariable=m4value, width=5)
        m4entry.grid(row=4, column=4, sticky=W, columnspan=1, padx=5)

        # Fill in row 5
        Label(frameBg, text='CC', bg = self.bg
              ).grid(row=5, column=0, sticky=E, columnspan=1, padx=5)
        ccvalue = StringVar(top)
        self.ccentry = Entry(frameBg, textvariable=ccvalue, width=3)
        self.ccentry.grid(row=5, column=1, sticky=W, columnspan=1, padx=5)
        l5value = StringVar(top)
        self.l5entry = Entry(frameBg, textvariable=l5value, width=12)
        self.l5entry.grid(row=5, column=2, sticky=W, columnspan=2, padx=5)
        m5value = StringVar(top)
        m5entry = Entry(frameBg, textvariable=m5value, width=5)
        m5entry.grid(row=5, column=4, sticky=W, columnspan=1, padx=5)

        button = Button(frameBg, text='Proceed', command=self.Proceed, takefocus=0)
        button.grid(row=5, column=11, sticky=N+EW, rowspan=2, columnspan=1, padx=5)

        # Fill in row 6
        Label(frameBg, text='SP', bg = self.bg
              ).grid(row=6, column=0, sticky=E, columnspan=1, padx=5)
        spvalue = StringVar(top)
        spentry = Entry(frameBg, textvariable=spvalue, width=5)
        spentry.grid(row=6, column=1, sticky=W, columnspan=1, padx=5)
        l6value = StringVar(top)
        l6entry = Entry(frameBg, textvariable=l6value, width=12)
        l6entry.grid(row=6, column=2, sticky=W, columnspan=2, padx=5)
        m6value = StringVar(top)
        m6entry = Entry(frameBg, textvariable=m6value, width=5)
        m6entry.grid(row=6, column=4, sticky=W, columnspan=1, padx=5)

        # Make lists of entry widgets for <Return> event handling
        # Register entry lists
        self.reglist = [pcentry, yentry, self.xentry, self.aentry,
                         self.bentry, self.ccentry, spentry]

        for entry in self.reglist:
            entry.bind('<Return>', self.ChangeReg) # Change any register
            entry.bind('<FocusOut>', self.ChangeReg)
            entry.bind('<FocusIn>', self.ReFocus)

        self.regentryl = [{'regvalue':pcvalue, 'regsize':2, 'regcmd':'RM P', 'replylen':56},
                          {'regvalue':yvalue, 'regsize':2,  'regcmd':'RM Y', 'replylen':56},
                          {'regvalue':xvalue, 'regsize':2,  'regcmd':'RM X', 'replylen':56},
                          {'regvalue':avalue, 'regsize':1,  'regcmd':'RM A', 'replylen':54},
                          {'regvalue':bvalue, 'regsize':1,  'regcmd':'RM B', 'replylen':54},
                          {'regvalue':ccvalue, 'regsize':1,  'regcmd':'RM C', 'replylen':54},
                          {'regvalue':spvalue, 'regsize':2,  'regcmd':'RM S', 'replylen':56}]

        # Address entry lists
        self.addrlist = [l1entry, l2entry, l3entry, l4entry, self.l5entry, l6entry]
        for entry in self.addrlist:
            entry.bind('<Return>', self.ChangeAddr) # Display new memory location
            entry.bind('<FocusOut>', self.ChangeAddr)
            entry.bind('<FocusIn>', self.ReFocus)

        self.addentryl = [l1value, l2value, l3value, l4value, l5value, l6value]

        # Memory Entry and Value lists
        self.memlist = [m1entry, m2entry, m3entry, m4entry, m5entry, m6entry]
        for entry in self.memlist:
            entry.bind('<Return>', self.ChangeMem) # Change the contents of memory
            entry.bind('<FocusOut>', self.ChangeMem)
            entry.bind('<FocusIn>', self.ReFocus)

        self.mementryl = [{'memvalue':m1value, 'memsize':1},
                          {'memvalue':m2value, 'memsize':1},
                          {'memvalue':m3value, 'memsize':1},
                          {'memvalue':m4value, 'memsize':2},
                          {'memvalue':m5value, 'memsize':2},
                          {'memvalue':m6value, 'memsize':2}]

        # Fill in row 7
        Label(frameBg, text = 'Break Points', bg = self.bg
              ).grid(row=7, column=0, sticky=S+EW, columnspan=5, pady=5)

        button = Button(frameBg, text='Update', takefocus=0)
        button.grid(row=7, column=5, sticky=EW, columnspan=1, padx=5)
        button.bind('<ButtonRelease-1>', self.Update)
        button = Button(frameBg, text='Memory Dump', command=self.MemoryDump, takefocus=0)
        button.grid(row=7, column=6, sticky=EW, columnspan=1, padx=5)

        button = Button(frameBg, text='Go', command=self.Go, takefocus=0)
        button.grid(row=7, column=11, sticky=N+EW, rowspan=2, columnspan=1, padx=5)

        # Fill in row 8
        frameBrEntry = Frame(frameBg, bg=self.bg)
        frameBrEntry.grid(row=8, column=0, sticky=NE+SW, rowspan=2, columnspan=5, padx=0)
        Label(frameBrEntry, text='BP1', bg = self.bg
              ).grid(row=0, column=0, sticky=SE, columnspan=1, padx=0)
        bp1value = StringVar(top)
        bp1entry = Entry(frameBrEntry, textvariable=bp1value, width=12)
        bp1entry.grid(row=0, column=1, sticky=SW, columnspan=1, padx=5)
        Label(frameBrEntry, text='BP2', bg = self.bg
              ).grid(row=0, column=2, sticky=SE, columnspan=1, padx=0)
        bp2value = StringVar(top)
        bp2entry = Entry(frameBrEntry, textvariable=bp2value, width=12)
        bp2entry.grid(row=0, column=3, sticky=SW, columnspan=1, padx=5)

        button = Button(frameBg, text='DnLoad', command=self.Load, takefocus=0)
        button.grid(row=8, column=5, sticky=EW, rowspan=2, columnspan=1, padx=5)

        self.s19value = StringVar(top)
        self.s19entry = s19entry = Entry(frameBg, textvariable=self.s19value, width=20)
        s19entry.grid(row=8, column=6, sticky=EW, rowspan=2, columnspan=4, padx=5)
        button = Button(frameBg, text='Browse', command=self.Browse, takefocus=0)
        button.grid(row=8, column=10, sticky=EW, rowspan=2, columnspan=1, padx=5, pady=16)

        # Fill in row 9
        Label(frameBrEntry, text='BP3', bg = self.bg
              ).grid(row=1, column=0, sticky=E, columnspan=1, padx=0)
        bp3value = StringVar(top)
        bp3entry = Entry(frameBrEntry, textvariable=bp3value, width=12)
        bp3entry.grid(row=1, column=1, sticky=NW, columnspan=2, padx=5)
        Label(frameBrEntry, text='BP4', bg = self.bg
              ).grid(row=1, column=2, sticky=E, columnspan=1, padx=0)
        bp4value = StringVar(top)
        bp4entry = Entry(frameBrEntry, textvariable=bp4value, width=12)
        bp4entry.grid(row=1, column=3, sticky=NW, columnspan=2, padx=5)

        # Breakpoint Entry and Value lists
        self.bplist = [bp1entry, bp2entry, bp3entry, bp4entry]

        for entry in self.bplist:
            entry.bind('<Return>', self.ChangeBP) # Change a breakpoint
            entry.bind('<FocusOut>', self.ChangeBP)
            entry.bind('<FocusIn>', self.ReFocus)

        self.brkentryl = [bp1value, bp2value, bp3value, bp4value]
        self.breakpoints = ['', '', '', '']     # List of actually set breakpoints

        # Fill in row 10
        Frame(frameBg, borderwidth=1, relief=SUNKEN,
              height=2, bg=self.bg).grid(row=10, column=0, sticky=EW,
                                         columnspan=12, padx=0, pady=0)  # Draw the separator

        # Fill in row 11
        self.buttonGUI = button = Button(frameBg, text='GUI Only',
                                         command=self.IGToggle, takefocus=0)
        button.grid(row=11, column=0, sticky=N, columnspan=3, padx=5, pady=3)
        self.termlable = Label(frameBg, text='HC11 SCI - Terminal', bg = self.bg)
        self.termlable.grid(row=11, column=5, sticky=S+EW, columnspan=3, padx=1)
        Label(frameBg, text='Baud Rate', bg = self.bg
              ).grid(row=11, column=10, sticky=SE, columnspan=1, padx=5)
        self.baudrateValue = IntVar(top)
        self.baudrateValue.set(self.baudrate)
        brentry = Entry(frameBg, textvariable=self.baudrateValue, width=6)
        brentry.grid(row=11, column=11, sticky=SW, columnspan=1, padx=5)

        brentry.bind('<Return>', self.ChangeBaud)
#        brentry.bind('<FocusOut>', self.ChangeBaud)
        brentry.bind('<FocusIn>', self.ReFocus)

        self.frameText = frameText = Frame(top, relief=FLAT)
        self.scrollbarView = Scrollbar(frameText, orient=VERTICAL,
                                       takefocus=FALSE, highlightthickness=0)
        self.textView = textView = Text(frameText, wrap=WORD, highlightthickness=0, takefocus=0,
                                        fg=self.fg, bg=self.bg, bd=2, font=('Courier', 10))
        self.scrollbarView.config(command=textView.yview)
        textView.config(yscrollcommand=self.scrollbarView.set)
        self.scrollbarView.pack(side=RIGHT, fill=Y)
        textView.pack(side=LEFT, expand=TRUE, fill=BOTH)
        frameText.grid(row=1, column=0, sticky=EW)

        textView.bind('<Control-KeyPress>', self.OnTermControlKeyPress)
        textView.bind('<KeyPress>', self.OnTermKeyPress)
        textView.bind('<Return>', self.OnTermReturnKey)
        textView.bind('<Tab>', self.OnTermTabKey)

    def notresponding(self):
        message = "The targer is not responding"
        m = tkMessageBox.Message(
            title = "Communications Timeout",
            message = message,
            icon = tkMessageBox.WARNING,
            type = tkMessageBox.OK,
            master = self.top)
        return m.show()

    # Act as output file
    def write(self, s, tags=(), mark="insert"):
        self.textView.insert(mark, s, tags)
        self.textView.see(mark)
        self.textView.update()

    def writelines(self, l):
        map(self.write, l)

###################################################################

    def ReFocus(self, event):
        event.widget.select_range(0, END)

    def OnTermControlKeyPress(self, event):
        keysym = event.keysym
        key = ''
        if keysym == 'x':
            key = chr(0x18)
        elif keysym == 'w':
            key = chr(0x17)
        elif keysym == 'd':
            key = chr(0x04)
        elif keysym == 'c':
            key = chr(0x03)
        elif keysym == 'b':
            key = chr(0x02)
        elif keysym == 'a':
            key = chr(0x01)
        if key:
            self.SendChars(key)
        if keysym < 'c':
            return 'break'

    def OnTermTabKey(self, event):
        self.reglist[0].focus_set()
        return 'break'

    def OnTermReturnKey(self, event):
        # move insertion point to the end?
        # maybe for all key presses?
        self.OnTermKeyPress(event)

    def OnTermKeyPress(self, event):
        keysym = event.keysym
        if keysym not in notkeys:
#            print event.char, keysym, event.keycode,
            key = event.char
            if keysym == 'Delete':
#                print event.keycode
                key = chr(0x7F)
            self.SendChars(key)
        if keysym != 'BackSpace':
            return 'break'

    def AsmListDoubleClick(self, event):
        asmlist = event.widget
        asmline = asmlist.get(asmlist.curselection())
        self.regentryl[0]['regvalue'].set(asmline[:4])
        self.ChangeReg()

    def ChangeBaud(self, event):
        oldfw = self.top.focused_widget
        self.top.focused_widget = event.widget
        baud = self.baudrateValue.get()
        if baud == self.baudrate:
            return
        if baud > 9600 or baud % 300:
            print >> self, 'Bad baud rate!'
            return
        multi = 9600//baud
        divis = 0
        while 1:
            multi = multi >> 1
            if not multi:
                break
            divis += 1
        cmd = 'MM 102B' + chr(13)   # Write target memory
        arg = '3' + str(divis)
        self.Converse(cmd, expect=17, tofactor=.05)
        self.Converse(arg, expect=2)
        self.Converse(chr(13), expect=10, notify=False)
        self.tty.setbaudrate(baud)
        self.baudrate  = baud
        self.top.focused_widget = oldfw

    def MemoryDump(self, event=None):
#      '0000 AA 55 38 72 86 AA 97 00 C6 55 D7 01 3D DD 02 39  U8r     U  =  9        '
#       78 characters X 9 lines = 702 + newline and prompt = 704
        addr = self.addentryl[0].get() or '0000'
        try:
            addr = self.constxref[addr]
        except KeyError:
            try:
                a = int(addr, 16)
                if a > 0xFFFF:
                    raise ValueError
                addr = hex(a)[2:].upper().zfill(4)
            except ValueError:
                addr = '0000'
        replystring = self.Converse('MD ' + addr + chr(13), expect=712, tofactor=.0032)
        try:
            reply = replystring.split('\n')[2:-1]
        except IndexError:
            return
        self.write('\n'.join(reply) + '\n' * 2)

    def Call(self, event=None):
        replystring = self.Converse('CALL' + chr(13), expect=69)
        try:
            reply = replystring.split('\n')[-2]
        except IndexError:
            return
        self.Update(reply = reply)

    def Step(self, event=None):
#       t

#       TEST                 P-E365 Y-0005 X-00A6 A-0A B-09 C-90 S-003E '
#       >
        self.SendChars('T' + chr(13), 69)
        try:
            reply = self.replystring.split('\n')[-2]
        except IndexError:
            print 'Got a bad reply from Trace'
            return
        try:
            activeline = self.asmAddrList.index(self.regentryl[0]['regvalue'].get())
        except ValueError:
            print 'ValueError'
            pass
        else:   # Clear the dis-assembly list line if the last address is in addr list
            self.disAsmList.select_clear(activeline)
        try:
            activeline = self.asmAddrList.index(reply[23:27])
        except ValueError:
            pass
        else:
            self.disAsmList.select_set(activeline)
            self.disAsmList.see(activeline)
        self.Update(reply = reply[21:-1])

    def Proceed(self, event=None):
        self.SendChars('P' + chr(13), 2)

    def Go(self, event=None):
        self.SendChars('G' + chr(13), 2)

    def Load(self, event=None):
        fname = self.s19value.get()
        if fname:
            self.SendChars('LOAD T' + chr(13), 7)
            try:
                tmpfile = open(fname)
                self.top.config(cursor="watch")
                for line in tmpfile:
                    self.root.update()
                    for char in line:
                        self.SendChars(char)
            except IOError:
                return
            tmpfile.close()
            rc = self.CollectReply(18, tofactor=.1, nosend=True)     # The promt will arrive
            if not rc:
                return
            if rc > 6:
                self.top.bell()
                if rc == 11:
                    self.SendChars(chr(13), 9)
                    print >> self, 'Download Error: Read Only Memory'
                elif rc > 11:
                    print >> self, 'Download Error: bad checksum'
                else:
                    print >> self, 'Download Error: the target isn\'t ready'
                return
            cmd = 'BR -' + chr(13)  # Clear breakpoints
            self.SendChars(cmd, 29)
            self.breakpoints = ['', '', '', '']
            for entry in self.brkentryl:    # Clear all entry widgets
                entry.set('')
            for entry in self.addentryl:
                entry.set('')
            for entrydict in self.mementryl:
                entrydict['memvalue'].set('')
            for entrydict in self.regentryl:
                entrydict['regvalue'].set('')
                self.disAsmList.delete(0, END)
            self.labelxref = {}          # Dictionary of labels and their values
            self.constxref = {}          # Dictionary of constants and their values
            self.asmAddrList = []   # List of address in the loaded context
            fname = fname[:-4] + '.lst'
            tmpfile = open(fname)
            for i, line in enumerate(tmpfile):
                self.disAsmList.insert(i, line[6:-1])   # Instert line into list box
                self.asmAddrList.append(line[6:10])     # Append to list for address lookup
                self.root.update()
            tmpfile.close()
            fname = fname[:-4] + '.xrf'
            tmpfile = open(fname)
            tmpfile.readline()      # Skip Labels header lines of the file
            tmpfile.readline()      # .readline() can be used before .next()
            tmpfile.readline()
            for line in tmpfile:    # for loops call .next() repeatedly
                if line[0] == ' ':
                    continue
                lablinelist = line.split()
                try:
                    self.labelxref[lablinelist[0]] = lablinelist[1][1:]
                except IndexError:
                    break
                self.root.update()
            tmpfile.next()          # Skip Constants header lines of the file
            tmpfile.next()          # .readline() can NOT be used after .next()
            tmpfile.next()          # because of look-ahead buffering
            for line in tmpfile:
                if line[0] == ' ':
                    continue
                lablinelist = line.split()
                try:
                    self.constxref[lablinelist[0]] = lablinelist[1][1:]
                except IndexError:
                    break
                self.root.update()
            tmpfile.close()
            self.top.config(cursor="")
            self.top.bell()
        else:
            self.Browse()

    def Browse(self, event=None):
        srecname = tkFileDialog.Open(parent=self.top,
                                     filetypes = [("S records", "*.s19", "TEXT")]).show()
        self.s19value.set(srecname)
        return srecname

    def IGToggle(self, event=None):
        if not self.gui_only:
            self.top.resizable(height=TRUE, width=TRUE)    # unlock the size
            self.frameText.grid_remove()
            self.top.resizable(height=FALSE, width=FALSE)    # then lock the size
            self.buttonGUI.config(text="Interactive")
            self.termlable.config(text="")
        else:
            self.top.resizable(height=TRUE, width=TRUE)    # unlock the size
            self.frameText.grid()
            self.top.resizable(height=FALSE, width=FALSE)    # then lock the size
            self.buttonGUI.config(text="GUI Only")
            self.termlable.config(text="HC11 SCI - Terminal")
        self.gui_only = not self.gui_only
        self.top.wm_deiconify()

    def ChangeBP(self, event):
#       br 0001
#
#
#       0001 0000 0000 0000 
#       >
        thisentry = event.widget
        teindex = self.bplist.index(thisentry)
        chars = self.brkentryl[teindex].get()
        oldbp = self.breakpoints[teindex]
        if not chars:
            if oldbp:
                self.RemoveBP(oldbp, teindex)
        else:
            try:
                chars = self.labelxref[chars]
            except KeyError:
                try:
                    a = int(chars, 16)
                    if a > 0xFFFF:
                        raise ValueError
                    chars = hex(a)[2:].upper().zfill(4)
                    self.brkentryl[teindex].set(chars)
                except ValueError:
                    thisentry.bell()
                    thisentry.focus_set()
                    thisentry.select_range(0, END)
                    return "break"
            if oldbp:
                self.RemoveBP(oldbp, teindex)
            cmd = 'BR ' + chars + chr(13)
            self.SendChars(cmd, 31)
            self.breakpoints[teindex] = chars
        if event.type == '10':
            return
        if teindex == 3:
            teindex = 0
        else:
            teindex += 1
        self.bplist[teindex].focus_set()
        self.bplist[teindex].select_range(0, END)

    def RemoveBP(self, oldbp, index):
        cmd = 'BR -' + oldbp + chr(13)
        self.SendChars(cmd, 33)
        self.breakpoints[index] = ''

    def ChangeReg(self, event=None):
        if event:
            thisentry = event.widget
            teindex = self.reglist.index(thisentry)
        else:
            thisentry = self.reglist[0]
            teindex = 0
        tedict = self.regentryl[teindex]
        chars = tedict['regvalue'].get()
        if chars:
            size = tedict['regsize']
            try:
                a = int(chars, 16)
                if a > 256**(tedict['regsize'])-1:
                    raise ValueError
                chars = hex(a)[2:].upper().zfill(size * 2)
                tedict['regvalue'].set(chars)
            except ValueError:
                thisentry.bell()
                thisentry.select_range(0, END)
                return "break"
            if teindex == 0:
                self.disAsmList.select_clear(0, END)
                try:
                    activeline = self.asmAddrList.index(chars)
                except ValueError:
                    pass
                else:
                    self.disAsmList.select_set(activeline)
                    self.disAsmList.see(activeline)
            cmd = tedict['regcmd'] + chr(13)
            replen = tedict['replylen']
            self.SendChars(cmd, replen)
            self.SendChars(chars + chr(13), 4 + size * 2)
        if event:
            if event.type == '10':
                return
            if teindex == 6:
                teindex = 0
            else:
                teindex += 1
        else:
            teindex = 0
        self.reglist[teindex].focus_set()
        self.reglist[teindex].select_range(0, END)

    def Update(self, event=None, reply=''):
#       P-FFFF Y-FFFF X-FFFF A-FF B-FF C-D0 S-0047 on line 2
#       P-FFFF on line 3
        if event:
            cmd = 'RM' + chr(13) * 2
            self.SendChars(cmd, replylen=58, tofactor=.0032)
            try:
                reply = self.replystring.split('\n')[1]
            except IndexError:
                return
        j = 2
        for i in range(len(self.regentryl)):
            regdict = self.regentryl[i]
            k = regdict['regsize'] * 2
            regdict['regvalue'].set(reply[j:j + k])
            j += k + 3
        for memdict in self.mementryl:
            try:
                addr = memdict['address']
            except KeyError:
                continue
            else:
                size = memdict['memsize']
                self.SendChars('MM ' + addr + chr(13), 17, .05)    # Read target memory
                value = self.replystring[-3:-1]
                if size == 2:
                    self.SendChars(' ', 4, .05)
                    value += self.replystring[-3:-1]
                memdict['memvalue'].set(value)
                self.SendChars(chr(13), 4)

    def ChangeAddr(self, event):
#       'mm
#
#       0005 06 '' FF ''
#
#       >'
        thisentry = event.widget
        teindex = self.addrlist.index(thisentry)
        memdict = self.mementryl[teindex]
        chars = self.addentryl[teindex].get()
        if chars:
            try:
                chars = self.constxref[chars]
            except KeyError:
                try:
                    a = int(chars, 16)
                    if a > 0xFFFF:
                        raise ValueError
                    chars = hex(a)[2:].upper().zfill(4)
                    self.addentryl[teindex].set(chars)
                except ValueError:
                    memdict['memvalue'].set('')
                    memdict.pop('address', 0)    # We don't care if it didn't exist.
                    thisentry.bell()
                    thisentry.focus_set()
                    thisentry.select_range(0, END)
                    return "break"
            memdict['address'] = chars
            size = memdict['memsize']
            self.SendChars('MM ' + chars + chr(13), 17, tofactor=.05)    # Read target memory
            value = self.replystring[-3:-1]
            if size == 2:
                self.SendChars(' ', 4, tofactor=.05)
                value += self.replystring[-3:-1]
            memdict['memvalue'].set(value)
            self.SendChars(chr(13), 4)
        else:
            memdict['memvalue'].set('')
            memdict.pop('address', 0)    # We don't care if it didn't exist.
        if event.type == '10':
            return
        if teindex == 5:
            teindex = 0
        else:
            teindex += 1
        self.addrlist[teindex].focus_set()
        self.addrlist[teindex].select_range(0, END)

    def ChangeMem(self, event):
#       'mm                     3
#                               1
#       0005 06 '' FF ''        12
#                               1
#                               1
#       rom-                    5
#       >'                      1
        thisentry = event.widget
        teindex = self.memlist.index(thisentry)
        tedict = self.mementryl[teindex]
        chars = tedict['memvalue'].get()
        if chars:
            try:
                addr = tedict['address']
            except KeyError:
                if event.type == '10':
                    return
                thisentry.bell()
                self.addrlist[teindex].focus_set()
                self.addrlist[teindex].select_range(0, END)
                return
            try:
                a = int(chars, 16)
                size = tedict['memsize']
                if a > 256**size -1:
                    raise ValueError
                chars = hex(a)[2:].upper().zfill(size * 2)
                tedict['memvalue'].set(chars)
            except ValueError:
                thisentry.bell()
                thisentry.select_range(0, END)
                return "break"
            cmd = 'MM ' + addr + chr(13)            # Write target memory
            self.SendChars(cmd, 17, tofactor=.01)
            self.SendChars(chars[:2] + ' ', replylen=0)      # The promp won't be sent back
            rc = self.CollectReply(11, tofactor=.0128, nosend=True)
            if rc > 6:
                thisentry.bell()
                tedict['memvalue'].set('rom-')
                thisentry.select_range(0, END)
                self.SendChars(chr(13), 3)
                return
            if size == 2:
                self.SendChars(chars[2:4], replylen=0)
            self.SendChars(chr(13), replylen=0)
            self.CollectReply(5, tofactor=.032, nosend=True)
        if event.type == '10':
            return
        if teindex == 5:
            teindex = 0
        else:
            teindex += 1
        self.memlist[teindex].focus_set()
        self.memlist[teindex].select_range(0, END)

    def SendChars(self, chars, replylen=0, tofactor=.04):
        self.catchecho = replylen
        self.replycount = 0
        self.replystring = ''
        self.tty.write(chars)
        if replylen:
            self.CollectReply(replylen, tofactor)

    def CollectReply(self, replylen, tofactor=.04, nosend=False):
        if nosend:
            self.catchecho = replylen
            self.replycount = 0
            self.replystring = ''
        starttime = time()
        while self.replycount < self.catchecho:
            self.PollQueue()
            if time() > starttime + (self.catchecho *
                                    (tofactor *
                                    (9600//self.baudrate))):
                if not nosend:
                    self.notresponding()
                    self.top.wakeup()
                break
        return self.replycount

    def Converse(self, chars, expect=0, tofactor=.04, notify=True):
        self.root.after_cancel(self.PCID)
        self.root.update()
        self.tty.write(chars)
        replystring = ''
        timeout = time() + (expect * (tofactor * (9600//self.baudrate)))
        while expect:
            done, reply = self.PollQueue(expect=expect)
            expect -= len(reply)
            replystring += reply
            if done or expect <= 0:
                break
            if time() > timeout:
                if notify:
                    self.notresponding()
                    self.top.wakeup()
                break
        self.PeriodicCall()
        return replystring

    def PeriodicCall(self):
        if self.running:
            self.PCID = self.root.after(25, self.PeriodicCall)
            self.PollQueue()

    def PollQueue(self, expect=0):
        outstring = ''
        count = 0
        while self.queue.qsize():
            data = self.queue.get(.004) # Get character or string w/ .004S timeout
            for char in data:
                if ord(char) <= 12:
                    continue
                if ord(char) == 13:
                    char = '\n'
                outstring += char
                if count < expect:
                    if char == '>':
                        return True, outstring
                    count += 1
                    return (count == expect), outstring
        if outstring:
            self.write(outstring)
        return False, ''

    def close(self, event=None):
        self.running = False
        self.root.after_cancel(self.PCID)
        self.tty.close()
        if self.flist:
            self.flist.debugger = None
        self.queue = None
        self.top.destroy()


# A basic synchronous COM port file-like object
class SerialTTY:
    def __init__(self, master, port, baud):
        self.master = master
        self.port = port
        self.running = master.running
        self.eventStop = CreateEvent(None, 0, 0, None)
        self.overlapped = OVERLAPPED()
        self.overlapped.hEvent = CreateEvent(None, 1, 0, None)
        self.handle = CreateFile(port,
                                 win32con.GENERIC_READ | win32con.GENERIC_WRITE,
                                 0, # exclusive access
                                 None, win32con.OPEN_EXISTING,
                                 win32con.FILE_ATTRIBUTE_NORMAL | win32con.FILE_FLAG_OVERLAPPED,
                                 None)
        SetCommMask(self.handle, EV_RXCHAR) # Tell the port we want a notification on each char.
        SetupComm(self.handle, 4096, 4096)  # Setup a 4k buffer
        # Remove anything that was there
        PurgeComm(self.handle, PURGE_TXABORT | PURGE_RXABORT | PURGE_TXCLEAR | PURGE_RXCLEAR )
        timeouts = 0xFFFFFF, 0, 1000, 0, 1000   # ReadIntervalTimeout
        SetCommTimeouts(self.handle, timeouts)  # ReadTotalTimeoutMultiplier, ReadTotalTimeoutConstant
        dcb = GetCommState( self.handle )       # WriteTotalTimeoutMultiplier, WriteTotalTimeoutConstant
        dcb.BaudRate = baud       # Setup the connection info.
        dcb.ByteSize = 8
        dcb.Parity = NOPARITY
        dcb.StopBits = ONESTOPBIT
        SetCommState(self.handle, dcb)
        print "Connected to %s at %s baud" % (port, dcb.BaudRate)

    def setbaudrate(self, baud):
        if baud > 9600 or baud % 300:
            print "Bad baud rate!"
            return
        PurgeComm(self.handle, PURGE_TXABORT | PURGE_RXABORT | PURGE_TXCLEAR | PURGE_RXCLEAR )
        dcb = GetCommState( self.handle )
        dcb.BaudRate = baud
        SetCommState(self.handle, dcb)
        print "Connected to %s at %s baud" % (self.port, dcb.BaudRate)

    def write(self, chars):
        try:
            WriteFile(self.handle, chars, self.overlapped)
            # Wait for the write to complete.
            WaitForSingleObject(self.overlapped.hEvent, INFINITE)
        finally:
            pass

    def read(self):
        overlapped = OVERLAPPED()
        overlapped.hEvent = CreateEvent(None, 1, 0, None)
        while self.running:
            rc, mask = WaitCommEvent(self.handle, overlapped)
            if rc == 0: # Character already ready!
                SetEvent(overlapped.hEvent)
            rc = WaitForMultipleObjects([overlapped.hEvent, self.eventStop], 0, INFINITE)
            if rc == WAIT_OBJECT_0:
                # Some input - read and print it
                flags, comstat = ClearCommError( self.handle )
                rc, data = ReadFile(self.handle, comstat.cbInQue, overlapped)
                WaitForSingleObject(overlapped.hEvent, INFINITE)
                self.master.queue.put(str(data))
            else:
                break   # Stop the thread!
            
    def close(self):
        SetEvent(self.eventStop)
        self.handle = CloseHandle(self.handle)

if __name__ == '__main__':
    # test the dialog
    root = Tk()
    dbug = Debugger(root)
    root.mainloop()
    root.destroy()
